import logging

from krrood.entity_query_language.entity import entity, let, in_, infer
from krrood.entity_query_language.symbolic import rule_mode
from numpy.ma.testutils import (
    assert_equal,
)  # You could replace this with numpy's regular assert for better compatibility

from semantic_digital_twin.reasoning.world_reasoner import WorldReasoner
from semantic_digital_twin.testing import *
from semantic_digital_twin.semantic_annotations.semantic_annotations import *

try:
    from ripple_down_rules.user_interface.gui import RDRCaseViewer
    from PyQt6.QtWidgets import QApplication
except ImportError as e:
    logging.debug(e)
    QApplication = None
    RDRCaseViewer = None

try:
    from semantic_digital_twin.reasoning.world_rdr import world_rdr
except ImportError:
    world_rdr = None


@dataclass(eq=False)
class TestSemanticAnnotation(SemanticAnnotation):
    """
    A Generic semantic annotation for multiple bodies.
    """

    _private_entity: KinematicStructureEntity = field(default=None)
    entity_list: List[KinematicStructureEntity] = field(
        default_factory=list, hash=False
    )
    semantic_annotations: List[SemanticAnnotation] = field(
        default_factory=list, hash=False
    )
    root_entity_1: KinematicStructureEntity = field(default=None)
    root_entity_2: KinematicStructureEntity = field(default=None)
    tip_entity_1: KinematicStructureEntity = field(default=None)
    tip_entity_2: KinematicStructureEntity = field(default=None)

    def add_entity(self, body: KinematicStructureEntity):
        self.entity_list.append(body)
        body._semantic_annotations.add(self)

    def add_semantic_annotation(self, semantic_annotation: SemanticAnnotation):
        self.semantic_annotations.append(semantic_annotation)
        semantic_annotation._semantic_annotations.add(self)

    @property
    def chain(self) -> list[KinematicStructureEntity]:
        """
        Returns itself as a kinematic chain.
        """
        return self._world.compute_chain_of_kinematic_structure_entities(
            self.root_entity_1, self.tip_entity_1
        )

    @property
    def _private_chain(self) -> list[KinematicStructureEntity]:
        """
        Private chain computation.
        """
        return self._world.compute_chain_of_kinematic_structure_entities(
            self.root_entity_2, self.tip_entity_2
        )


def test_semantic_annotation_hash(apartment_world):
    semantic_annotation1 = Handle(body=apartment_world.bodies[0])
    with apartment_world.modify_world():
        apartment_world.add_semantic_annotation(semantic_annotation1)
    assert hash(semantic_annotation1) == hash((Handle, apartment_world.bodies[0].name))

    semantic_annotation2 = Handle(body=apartment_world.bodies[0])
    assert semantic_annotation1 == semantic_annotation2


def test_aggregate_bodies(kitchen_world):
    world_semantic_annotation = TestSemanticAnnotation(_world=kitchen_world)

    # Test bodies added to a private dataclass field are not aggregated
    world_semantic_annotation._private_entity = (
        kitchen_world.kinematic_structure_entities[0]
    )

    # Test aggregation of bodies added in custom properties
    world_semantic_annotation.root_entity_1 = (
        kitchen_world.kinematic_structure_entities[1]
    )
    world_semantic_annotation.tip_entity_1 = kitchen_world.kinematic_structure_entities[
        4
    ]

    # Test aggregation of normal dataclass field
    body_subset = kitchen_world.kinematic_structure_entities[5:10]
    [world_semantic_annotation.add_entity(body) for body in body_subset]

    # Test aggregation of bodies in a new as well as a nested semantic annotation
    semantic_annotation1 = TestSemanticAnnotation()
    semantic_annotation1_subset = kitchen_world.kinematic_structure_entities[10:18]
    [semantic_annotation1.add_entity(body) for body in semantic_annotation1_subset]

    semantic_annotation2 = TestSemanticAnnotation()
    semantic_annotation2_subset = kitchen_world.kinematic_structure_entities[20:]
    [semantic_annotation2.add_entity(body) for body in semantic_annotation2_subset]

    semantic_annotation1.add_semantic_annotation(semantic_annotation2)
    world_semantic_annotation.add_semantic_annotation(semantic_annotation1)

    # Test that bodies added in a custom private property are not aggregated
    world_semantic_annotation.root_entity_2 = (
        kitchen_world.kinematic_structure_entities[18]
    )
    world_semantic_annotation.tip_entity_2 = kitchen_world.kinematic_structure_entities[
        20
    ]

    assert_equal(
        world_semantic_annotation.kinematic_structure_entities,
        set(kitchen_world.kinematic_structure_entities)
        - {
            kitchen_world.kinematic_structure_entities[0],
            kitchen_world.kinematic_structure_entities[19],
        },
    )


def test_handle_semantic_annotation_eql(apartment_world):
    with rule_mode():
        body = let(
            type_=Body,
            domain=apartment_world.bodies
        )
        query = infer(entity(Handle(body=body), in_("handle", body.name.name.lower())))

    handles = list(query.evaluate())
    assert len(handles) > 0


@pytest.mark.parametrize(
    "semantic_annotation_type, update_existing_semantic_annotations, scenario",
    [
        (Handle, False, None),
        (Container, False, None),
        (Drawer, False, None),
        (Cabinet, False, None),
        (Door, False, None),
    ],
)
def test_infer_apartment_semantic_annotation(
    semantic_annotation_type,
    update_existing_semantic_annotations,
    scenario,
    apartment_world,
):
    fit_rules_and_assert_semantic_annotations(
        apartment_world,
        semantic_annotation_type,
        update_existing_semantic_annotations,
        scenario,
    )


@pytest.mark.skipif(world_rdr is None, reason="requires world_rdr")
def test_generated_semantic_annotations(kitchen_world):
    found_semantic_annotations = world_rdr.classify(kitchen_world)[
        "semantic_annotations"
    ]
    drawer_container_names = [
        v.body.name.name for v in found_semantic_annotations if isinstance(v, Container)
    ]
    assert len(drawer_container_names) == 14


@pytest.mark.order("second_to_last")
def test_apartment_semantic_annotations(apartment_world):
    world_reasoner = WorldReasoner(apartment_world)
    world_reasoner.fit_semantic_annotations(
        [Handle, Container, Drawer, Cabinet],
        world_factory=lambda: apartment_world,
        scenario=None,
    )

    found_semantic_annotations = world_reasoner.infer_semantic_annotations()
    drawer_container_names = [
        v.body.name.name for v in found_semantic_annotations if isinstance(v, Container)
    ]
    assert len(drawer_container_names) == 19


def fit_rules_and_assert_semantic_annotations(
    world, semantic_annotation_type, update_existing_semantic_annotations, scenario
):
    world_reasoner = WorldReasoner(world)
    world_reasoner.fit_semantic_annotations(
        [semantic_annotation_type],
        update_existing_semantic_annotations=update_existing_semantic_annotations,
        world_factory=lambda: world,
        scenario=scenario,
    )

    found_semantic_annotations = world_reasoner.infer_semantic_annotations()
    assert any(
        isinstance(v, semantic_annotation_type) for v in found_semantic_annotations
    )


def test_semantic_annotation_serde_once(apartment_world):
    handle_body = apartment_world.bodies[0]
    door_body = apartment_world.bodies[1]

    handle = Handle(body=handle_body)
    door = Door(body=door_body, handle=handle)
    with apartment_world.modify_world():
        apartment_world.add_semantic_annotation(handle)
        apartment_world.add_semantic_annotation(door)

    door_se = door.to_json()
    door_de = Door.from_json(door_se)

    assert door == door_de
    assert type(door.handle) == type(door_de.handle)
    assert type(door.body) == type(door_de.body)


def test_semantic_annotation_serde_multiple(apartment_world):
    handle_body = apartment_world.bodies[0]
    door_body = apartment_world.bodies[1]

    handle = Handle(body=handle_body)
    door = Door(body=door_body, handle=handle)

    with apartment_world.modify_world():
        apartment_world.add_semantic_annotation(handle)
        apartment_world.add_semantic_annotation(door)

    door_se1 = door.to_json()
    door_de1 = Door.from_json(door_se1)

    assert door == door_de1
    assert type(door.handle) == type(door_de1.handle)
    assert type(door.body) == type(door_de1.body)

    door_se2 = door_de1.to_json()
    door_de2 = Door.from_json(door_se2)

    assert door == door_de2
    assert type(door.handle) == type(door_de2.handle)
    assert type(door.body) == type(door_de2.body)



from __future__ import absolute_import, annotations

from collections import OrderedDict
from functools import lru_cache
from typing import Dict, Tuple, TYPE_CHECKING

import numpy as np
import rustworkx.visit

from ..datastructures.prefixed_name import PrefixedName
from ..datastructures.types import NpMatrix4x4
from ..spatial_types import spatial_types as cas
from ..spatial_types.math import inverse_frame

from ..world_description.world_entity import Connection, KinematicStructureEntity

if TYPE_CHECKING:
    from ..world import World


class ForwardKinematicsVisitor(rustworkx.visit.DFSVisitor):
    """
    Visitor class for collection various forward kinematics expressions in a world model.

    This class is designed to traverse a world, compute the forward kinematics transformations in batches for different
    use cases.
    1. Efficient computation of forward kinematics between any bodies in the world.
    2. Efficient computation of forward kinematics for all bodies with collisions for updating collision checkers.
    3. Efficient computation of forward kinematics as position and quaternion, useful for ROS tf.
    """

    compiled_collision_fks: cas.CompiledFunction
    compiled_all_fks: cas.CompiledFunction

    forward_kinematics_for_all_bodies: np.ndarray
    """
    A 2D array containing the stacked forward kinematics expressions for all bodies in the world.
    Dimensions are ((number of bodies) * 4) x 4.
    They are computed in batch for efficiency.
    """
    body_name_to_forward_kinematics_idx: Dict[PrefixedName, int]
    """
    Given a body name, returns the index of the first row in `forward_kinematics_for_all_bodies` that corresponds to that body.
    """

    def __init__(self, world: World):
        self.world = world
        self.child_body_to_fk_expr: Dict[PrefixedName, cas.TransformationMatrix] = {
            self.world.root.name: cas.TransformationMatrix()
        }
        self.tf: Dict[Tuple[PrefixedName, PrefixedName], cas.Expression] = OrderedDict()

    def connection_call(self, edge: Tuple[int, int, Connection]):
        """
        Gathers forward kinematics expressions for a connection.
        """
        connection = edge[2]
        map_T_parent = self.child_body_to_fk_expr[connection.parent.name]
        self.child_body_to_fk_expr[connection.child.name] = map_T_parent.dot(
            connection.origin_expression
        )
        self.tf[(connection.parent.name, connection.child.name)] = (
            connection.origin_as_position_quaternion()
        )

    tree_edge = connection_call

    def compile_forward_kinematics(self) -> None:
        """
        Compiles forward kinematics expressions for fast evaluation.
        """
        all_fks = cas.Expression.vstack(
            [
                self.child_body_to_fk_expr[body.name]
                for body in self.world.kinematic_structure_entities
            ]
        )
        tf = cas.Expression.vstack([pose for pose in self.tf.values()])
        collision_fks = []
        for body in sorted(
            self.world.bodies_with_enabled_collision, key=lambda b: b.name
        ):
            if body == self.world.root:
                continue
            collision_fks.append(self.child_body_to_fk_expr[body.name])
        collision_fks = cas.Expression.vstack(collision_fks)
        params = [v.symbols.position for v in self.world.degrees_of_freedom]
        self.compiled_all_fks = all_fks.compile(parameters=[params])
        self.compiled_collision_fks = collision_fks.compile(parameters=[params])
        self.compiled_tf = tf.compile(parameters=[params])
        self.idx_start = {
            body.name: i * 4
            for i, body in enumerate(self.world.kinematic_structure_entities)
        }

    def recompute(self) -> None:
        """
        Clears cache and recomputes all forward kinematics. Should be called after a state update.
        """
        self.compute_forward_kinematics_np.cache_clear()
        self.subs = self.world.state.positions
        self.forward_kinematics_for_all_bodies = self.compiled_all_fks(self.subs)
        self.collision_fks = self.compiled_collision_fks(self.subs)

    def compute_tf(self) -> np.ndarray:
        """
        Computes a (number of bodies) x 7 matrix of forward kinematics in position/quaternion format.
        The rows are ordered by body name.
        The first 3 entries are position values, the last 4 entires are quaternion values in x, y, z, w order.

        This is not updated in 'recompute', because this functionality is only used with ROS.
        :return: A large matrix with all forward kinematics.
        """
        return self.compiled_tf(self.subs)

    @lru_cache(maxsize=None)
    def compute_forward_kinematics_np(
        self, root: KinematicStructureEntity, tip: KinematicStructureEntity
    ) -> NpMatrix4x4:
        """
        Computes the forward kinematics from the root body to the tip body, root_T_tip.

        This method computes the transformation matrix representing the pose of the
        tip body relative to the root body, expressed as a numpy ndarray.

        :param root: Root body for which the kinematics are computed.
        :param tip: Tip body to which the kinematics are computed.
        :return: Transformation matrix representing the relative pose of the tip body with respect to the root body.
        """
        root = root.name
        tip = tip.name
        root_is_world = root == self.world.root.name
        tip_is_world = tip == self.world.root.name

        if not tip_is_world:
            i = self.idx_start[tip]
            map_T_tip = self.forward_kinematics_for_all_bodies[i : i + 4]
            if root_is_world:
                return map_T_tip

        if not root_is_world:
            i = self.idx_start[root]
            map_T_root = self.forward_kinematics_for_all_bodies[i : i + 4]
            root_T_map = inverse_frame(map_T_root)
            if tip_is_world:
                return root_T_map

        if tip_is_world and root_is_world:
            return np.eye(4)

        return root_T_map @ map_T_tip



import os
import time
import unittest
from typing import Optional

import sqlalchemy
from krrood.ormatic.utils import drop_database
from semantic_digital_twin.semantic_annotations.semantic_annotations import Handle, Door
from sqlalchemy import select
from sqlalchemy.orm import Session

from semantic_digital_twin.adapters.ros.world_synchronizer import (
    StateSynchronizer,
    ModelReloadSynchronizer,
    ModelSynchronizer,
)
from semantic_digital_twin.adapters.urdf import URDFParser
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.orm.ormatic_interface import Base, WorldMappingDAO
from semantic_digital_twin.spatial_types import Vector3
from semantic_digital_twin.testing import rclpy_node
from semantic_digital_twin.utils import get_semantic_digital_twin_directory_root
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import (
    Connection6DoF,
    FixedConnection,
    PrismaticConnection,
)
from semantic_digital_twin.world_description.degree_of_freedom import DegreeOfFreedom
from semantic_digital_twin.world_description.world_entity import Body


def create_dummy_world(w: Optional[World] = None) -> World:
    if w is None:
        w = World()
    b1 = Body(name=PrefixedName("b1"))
    b2 = Body(name=PrefixedName("b2"))
    with w.modify_world():
        w.add_connection(Connection6DoF.create_with_dofs(parent=b1, child=b2, world=w))
    return w


def test_state_synchronization(rclpy_node):

    w1 = create_dummy_world()
    w2 = create_dummy_world()

    synchronizer_1 = StateSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = StateSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    # Allow time for publishers/subscribers to connect on unique topics
    time.sleep(0.2)

    w1.state.data[0, 0] = 1.0
    w1.notify_state_change()
    time.sleep(0.2)
    assert w1.state.data[0, 0] == 1.0
    assert w1.state.data[0, 0] == w2.state.data[0, 0]

    synchronizer_1.close()
    synchronizer_2.close()


def test_state_synchronization_world_model_change_after_init(rclpy_node):
    w1 = World()
    w2 = World()

    synchronizer_1 = StateSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    create_dummy_world(w1)
    create_dummy_world(w2)
    synchronizer_2 = StateSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    # Allow time for publishers/subscribers to connect on unique topics
    time.sleep(0.2)

    w1.state.data[0, 0] = 1.0
    w1.notify_state_change()
    time.sleep(0.2)
    assert w1.state.data[0, 0] == 1.0
    assert w1.state.data[0, 0] == w2.state.data[0, 0]

    synchronizer_1.close()
    synchronizer_2.close()


def test_model_reload(rclpy_node):

    engine = sqlalchemy.create_engine(
        "sqlite+pysqlite:///file::memory:?cache=shared",
        connect_args={"check_same_thread": False, "uri": True},
    )
    drop_database(engine)
    Base.metadata.create_all(engine)
    session_maker = sqlalchemy.orm.sessionmaker(bind=engine)
    session1 = session_maker()
    session2 = session_maker()

    w1 = create_dummy_world()
    w2 = World()

    synchronizer_1 = ModelReloadSynchronizer(
        rclpy_node,
        w1,
        session=session1,
    )
    synchronizer_2 = ModelReloadSynchronizer(
        rclpy_node,
        w2,
        session=session2,
    )

    synchronizer_1.publish_reload_model()
    time.sleep(1.0)
    assert len(w2.kinematic_structure_entities) == 2

    query = session1.scalars(select(WorldMappingDAO)).all()
    assert len(query) == 1
    assert w2.get_kinematic_structure_entity_by_name("b2")

    synchronizer_1.close()
    synchronizer_2.close()


def test_model_synchronization_body_only(rclpy_node):

    w1 = World(name="w1")
    w2 = World(name="w2")

    synchronizer_1 = ModelSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = ModelSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    with w1.modify_world():
        new_body = Body(name=PrefixedName("b3"))
        w1.add_kinematic_structure_entity(new_body)
        assert len(w1.kinematic_structure_entities) == 1

    time.sleep(0.2)
    assert len(w1.kinematic_structure_entities) == 1
    assert len(w2.kinematic_structure_entities) == 1

    assert w2.get_kinematic_structure_entity_by_name("b3")

    synchronizer_1.close()
    synchronizer_2.close()


def test_model_synchronization_creation_only(rclpy_node):

    w1 = World(name="w1")
    w2 = World(name="w2")

    synchronizer_1 = ModelSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = ModelSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    with w1.modify_world():
        b2 = Body(name=PrefixedName("b2"))
        w1.add_kinematic_structure_entity(b2)

        new_body = Body(name=PrefixedName("b3"))
        w1.add_kinematic_structure_entity(new_body)

        c = Connection6DoF.create_with_dofs(parent=b2, child=new_body, world=w1)
        w1.add_connection(c)
    time.sleep(0.1)
    assert len(w1.kinematic_structure_entities) == 2
    assert len(w2.kinematic_structure_entities) == 2
    assert len(w1.connections) == 1
    assert len(w2.connections) == 1

    synchronizer_1.close()
    synchronizer_2.close()


def test_model_synchronization_merge_full_world(rclpy_node):

    w1 = World(name="w1")
    w2 = World(name="w2")

    synchronizer_1 = ModelSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = ModelSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    pr2_world = URDFParser.from_file(
        os.path.join(
            get_semantic_digital_twin_directory_root(os.getcwd()),
            "resources",
            "urdf",
            "pr2_kinematic_tree.urdf",
        )
    ).parse()

    def wait_for_sync(timeout=3.0, interval=0.05):
        start = time.time()
        while time.time() - start < timeout:
            body_names_1 = [body.name for body in w1.kinematic_structure_entities]
            body_names_2 = [body.name for body in w2.kinematic_structure_entities]
            if body_names_1 == body_names_2:
                return body_names_1, body_names_2
            time.sleep(interval)

        body_names_1 = [body.name for body in w1.kinematic_structure_entities]
        body_names_2 = [body.name for body in w2.kinematic_structure_entities]
        return body_names_1, body_names_2

    with w1.modify_world():
        new_body = Body(name=PrefixedName("b3"))
        w1.add_kinematic_structure_entity(new_body)

    fixed_connection = FixedConnection(child=new_body, parent=pr2_world.root)
    w1.merge_world(pr2_world, fixed_connection)

    body_names_1, body_names_2 = wait_for_sync()

    assert body_names_1 == body_names_2
    assert len(w1.kinematic_structure_entities) == len(w2.kinematic_structure_entities)

    w1_connection_names = [c.name for c in w1.connections]
    w2_connection_names = [c.name for c in w2.connections]
    assert w1_connection_names == w2_connection_names
    assert len(w1.connections) == len(w2.connections)
    assert len(w2.degrees_of_freedom) == len(w1.degrees_of_freedom)

    synchronizer_1.close()
    synchronizer_2.close()


def test_callback_pausing(rclpy_node):

    w1 = World(name="w1")
    w2 = World(name="w2")

    model_synchronizer_1 = ModelSynchronizer(node=rclpy_node, world=w1)
    model_synchronizer_2 = ModelSynchronizer(node=rclpy_node, world=w2)
    state_synchronizer_1 = StateSynchronizer(node=rclpy_node, world=w1)
    state_synchronizer_2 = StateSynchronizer(node=rclpy_node, world=w2)

    model_synchronizer_2.pause()
    state_synchronizer_2.pause()
    assert model_synchronizer_2._is_paused
    assert state_synchronizer_2._is_paused

    with w1.modify_world():
        b2 = Body(name=PrefixedName("b2"))
        w1.add_kinematic_structure_entity(b2)

        new_body = Body(name=PrefixedName("b3"))
        w1.add_kinematic_structure_entity(new_body)

        c = Connection6DoF.create_with_dofs(parent=b2, child=new_body, world=w1)
        w1.add_connection(c)

    time.sleep(0.1)
    assert len(model_synchronizer_2.missed_messages) == 1
    assert len(w1.kinematic_structure_entities) == 2
    assert len(w2.kinematic_structure_entities) == 0
    assert len(w1.connections) == 1
    assert len(w2.connections) == 0

    state_synchronizer_2.resume()
    model_synchronizer_2.resume()
    state_synchronizer_2.apply_missed_messages()
    model_synchronizer_2.apply_missed_messages()

    time.sleep(0.1)
    assert len(w1.kinematic_structure_entities) == 2
    assert len(w2.kinematic_structure_entities) == 2
    assert len(w1.connections) == 1
    assert len(w2.connections) == 1


def test_ChangeDifHasHardwareInterface(rclpy_node):

    w1 = World(name="w1")
    w2 = World(name="w2")

    synchronizer_1 = ModelSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = ModelSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    with w1.modify_world():
        body1 = Body(name=PrefixedName("b1"))
        body2 = Body(name=PrefixedName("b2"))
        w1.add_kinematic_structure_entity(body1)
        w1.add_kinematic_structure_entity(body2)
        dof = DegreeOfFreedom(name=PrefixedName("dof"))
        w1.add_degree_of_freedom(dof)
        connection = PrismaticConnection(
            dof_name=dof.name, parent=body1, child=body2, axis=Vector3(1, 1, 1)
        )
        w1.add_connection(connection)
    assert len(w1.kinematic_structure_entities) == 2
    assert len(w1.connections) == 1

    time.sleep(0.2)
    assert len(w1.kinematic_structure_entities) == 2
    assert len(w2.kinematic_structure_entities) == 2
    assert len(w2.connections) == 1
    assert not w2.connections[0].dof.has_hardware_interface
    assert not w2.connections[0].dof.has_hardware_interface

    assert w2.get_kinematic_structure_entity_by_name("b2")

    with w1.modify_world():
        w1.set_dofs_has_hardware_interface(w1.degrees_of_freedom, True)

    time.sleep(0.2)
    assert w1.connections[0].dof.has_hardware_interface
    assert w2.connections[0].dof.has_hardware_interface

    synchronizer_1.close()
    synchronizer_2.close()


def test_semantic_annotation_modifications(rclpy_node):
    w1 = World(name="w1")
    w2 = World(name="w2")

    synchronizer_1 = ModelSynchronizer(
        node=rclpy_node,
        world=w1,
    )
    synchronizer_2 = ModelSynchronizer(
        node=rclpy_node,
        world=w2,
    )

    b1 = Body(name=PrefixedName("b1"))
    v1 = Handle(body=b1)
    v2 = Door(body=b1, handle=v1)

    with w1.modify_world():
        w1.add_body(b1)
        w1.add_semantic_annotation(v1)
        w1.add_semantic_annotation(v2)

    time.sleep(0.2)
    assert [sa.name for sa in w1.semantic_annotations] == [
        sa.name for sa in w2.semantic_annotations
    ]


if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import os
from dataclasses import dataclass

from typing_extensions import Optional, Tuple, Union, List, Dict
from urdf_parser_py import urdf as urdfpy

from ..datastructures.prefixed_name import PrefixedName
from ..exceptions import ParsingError
from ..spatial_types import spatial_types as cas
from ..spatial_types.derivatives import Derivatives, DerivativeMap
from ..spatial_types.spatial_types import TransformationMatrix, Vector3
from ..utils import (
    suppress_stdout_stderr,
    hacky_urdf_parser_fix,
    robot_name_from_urdf_string,
)
from ..world import World
from ..world_description.connections import (
    RevoluteConnection,
    PrismaticConnection,
    FixedConnection,
)
from ..world_description.degree_of_freedom import DegreeOfFreedom
from ..world_description.geometry import (
    Box,
    Sphere,
    Cylinder,
    FileMesh,
    Scale,
    Color,
)
from ..world_description.shape_collection import ShapeCollection
from ..world_description.world_entity import Body, Connection

connection_type_map = {  # 'unknown': JointType.UNKNOWN,
    "revolute": RevoluteConnection,
    "continuous": RevoluteConnection,
    "prismatic": PrismaticConnection,
    # 'floating': JointType.FLOATING,
    # 'planar': JointType.PLANAR,
    "fixed": FixedConnection,
}


def urdf_joint_to_limits(
    urdf_joint: urdfpy.Joint,
) -> Tuple[DerivativeMap[float], DerivativeMap[float]]:
    """
    Maps the URDF joint specifications to lower and upper joint limits, including
    position and velocity constraints. Mimics and safety controller parameters are
    also considered when determining the limits.

    :param urdf_joint: A URDF (Unified Robot Description Format) joint object
                       which contains the joint's type, limits, safety controller,
                       and mimic information.
    :return: A tuple containing two DerivativeMap objects, representing the lower
             and upper limits of the joint in terms of position and velocity.
    """
    lower_limits = DerivativeMap()
    upper_limits = DerivativeMap()
    limit = getattr(urdf_joint, "limit", None)
    if not urdf_joint.type == "continuous":
        lower = limit.lower if limit is not None else None
        upper = limit.upper if limit is not None else None

        safety_controller = getattr(urdf_joint, "safety_controller", None)
        lower = (
            max(safety_controller.soft_lower_limit, limit.lower)
            if safety_controller is not None and limit is not None
            else lower
        )
        upper = (
            min(safety_controller.soft_upper_limit, limit.upper)
            if safety_controller is not None and limit is not None
            else upper
        )

        lower_limits.position = lower
        upper_limits.position = upper

    velocity = getattr(limit, "velocity", None) if limit is not None else None
    lower_limits.velocity = -velocity if velocity is not None else None
    upper_limits.velocity = velocity if velocity is not None else None

    if urdf_joint.mimic is not None:
        multiplier = (
            urdf_joint.mimic.multiplier
            if urdf_joint.mimic.multiplier is not None
            else 1
        )
        offset = urdf_joint.mimic.offset if urdf_joint.mimic.offset is not None else 0

        for d2 in Derivatives.range(Derivatives.position, Derivatives.velocity):
            lower_limits.data[d2] -= offset
            upper_limits.data[d2] -= offset
            if multiplier < 0:
                upper_limits.data[d2], lower_limits.data[d2] = (
                    lower_limits.data[d2],
                    upper_limits.data[d2],
                )
            upper_limits.data[d2] /= multiplier
            lower_limits.data[d2] /= multiplier
    return lower_limits, upper_limits


@dataclass
class URDFParser:
    """
    Class to parse URDF files to worlds.
    Must set either urdf or file_path.
    """

    urdf: str
    """
    The URDF string.
    """

    prefix: Optional[str] = None
    """
    The prefix for every name used in this world.
    """

    package_resolver: Optional[Dict[str, str]] = None
    """
    The package resolver to use for resolving package paths in the URDF file. If ROS is installed, ROS will be used
     to resolve the paths, otherwise the package_resolver must be provided if the URDF file contains package paths.
     The key is the package name and the value is the path to the package. You can also set the environment variable
     `ROS_PACKAGE_PATH` to a colon-separated list of package paths, which will be used as the package resolver.
    """

    def __post_init__(self):
        self.urdf = hacky_urdf_parser_fix(self.urdf)
        self.parsed = urdfpy.URDF.from_xml_string(self.urdf)
        if self.prefix is None:
            self.prefix = robot_name_from_urdf_string(self.urdf)
        if self.package_resolver is None:
            package_paths = os.environ.get("ROS_PACKAGE_PATH", "").split(":")
            self.package_resolver = {
                os.path.basename(path): path
                for path in package_paths
                if os.path.exists(path)
            }

    @classmethod
    def from_file(cls, file_path: str, prefix: Optional[str] = None) -> URDFParser:
        if file_path is not None:
            with open(file_path, "r") as file:
                # Since parsing URDF causes a lot of warning messages which can't be deactivated, we suppress them
                with suppress_stdout_stderr():
                    urdf = file.read()
        return URDFParser(urdf=urdf, prefix=prefix)

    def parse(self) -> World:
        prefix = self.parsed.name
        links = [
            self.parse_link(link, PrefixedName(link.name, prefix))
            for link in self.parsed.links
        ]
        root = [link for link in links if link.name.name == self.parsed.get_root()][0]
        world = World()
        world.name = self.prefix
        with world.modify_world():
            world.add_kinematic_structure_entity(root)
            joints = []
            for joint in self.parsed.joints:
                parent = [link for link in links if link.name.name == joint.parent][0]
                child = [link for link in links if link.name.name == joint.child][0]
                parsed_joint = self.parse_joint(joint, parent, child, world, prefix)
                joints.append(parsed_joint)

            [world.add_connection(joint) for joint in joints]
            [world.add_kinematic_structure_entity(link) for link in links]

        return world

    def parse_joint(
        self, joint: urdfpy.Joint, parent: Body, child: Body, world: World, prefix: str
    ) -> Connection:
        """
        Parses a given URDF joint and creates a corresponding connection object.

        The function processes the provided joint data, extracting necessary
        information including translation offsets, rotation offsets, connection type,
        and relevant joint limits. It maps URDF joint types to predefined connection
        types and either retrieves or creates a degree of freedom (DOF) in the world
        context. It generates and returns a connection object representing the
        relationship between a parent and a child body.

        :param joint: The URDF joint to be parsed.
        :param parent: The parent body to be connected by the joint.
        :param child: The child body to be connected by the joint.
        :param world: The world instance containing degrees of freedom.
        :param prefix: The prefix for naming connections and DOFs.
        :return: A connection object representing the parsed joint.
        """
        connection_name = PrefixedName(joint.name, prefix)
        connection_type = connection_type_map.get(joint.type, Connection)
        translation_offset = getattr(joint.origin, "xyz", [0, 0, 0])
        rotation_offset = getattr(joint.origin, "rpy", [0, 0, 0])

        parent_T_connection = cas.TransformationMatrix.from_xyz_rpy(
            x=translation_offset[0],
            y=translation_offset[1],
            z=translation_offset[2],
            roll=rotation_offset[0],
            pitch=rotation_offset[1],
            yaw=rotation_offset[2],
        )
        if connection_type is FixedConnection:
            return FixedConnection(
                name=connection_name,
                parent=parent,
                child=child,
                parent_T_connection_expression=parent_T_connection,
            )

        lower_limits, upper_limits = urdf_joint_to_limits(joint)
        dof_name = connection_name
        multiplier = offset = None
        if joint.mimic:
            multiplier = (
                joint.mimic.multiplier if joint.mimic.multiplier is not None else 1
            )
            offset = joint.mimic.offset if joint.mimic.offset is not None else 0
            dof_name = PrefixedName(joint.mimic.joint, prefix)

        try:
            dof = world.get_degree_of_freedom_by_name(dof_name)
        except KeyError as e:
            dof = DegreeOfFreedom(
                name=dof_name,
                lower_limits=lower_limits,
                upper_limits=upper_limits,
            )
            world.add_degree_of_freedom(dof)

        assert joint.axis is not None, f"Joint axis is None for joint {joint.name}"
        result = connection_type(
            name=connection_name,
            parent=parent,
            child=child,
            parent_T_connection_expression=parent_T_connection,
            multiplier=multiplier,
            offset=offset,
            axis=Vector3(*map(int, joint.axis), reference_frame=parent),
            dof_name=dof.name,
        )
        return result

    def parse_link(self, link: urdfpy.Link, parent_frame: PrefixedName) -> Body:
        """
        Parses a URDF link to a link object.
        :param link: The URDF link to parse.
        :param parent_frame: The parent frame of the link, used for transformations of collisions and visuals.
        :return: The parsed link object.
        """
        name = PrefixedName(prefix=self.prefix, name=link.name)
        body = Body(name=name)
        visuals = self.parse_geometry(link.visuals, body)
        collisions = self.parse_geometry(link.collisions, body)
        body.visual = visuals
        body.collision = collisions
        return body

    def parse_geometry(
        self,
        geometry: Union[List[urdfpy.Collision], List[urdfpy.Visual]],
        body: Body,
    ) -> ShapeCollection:
        """
        Parses a URDF geometry to the corresponding shapes.
        :param geometry: The URDF geometry to parse either the collisions of visuals.'
        :param body: The body of the geometry, used for back referencing.
        :return: A List of shapes corresponding to the URDF geometry.
        """
        res = []
        material_dict = dict(
            zip(
                [material.name for material in self.parsed.materials],
                [
                    material.color.rgba if material.color else None
                    for material in self.parsed.materials
                ],
            )
        )
        for i, geom in enumerate(geometry):
            params = (*(geom.origin.xyz + geom.origin.rpy),) if geom.origin else ()
            origin_transform = TransformationMatrix.from_xyz_rpy(
                *params, reference_frame=body
            )
            if isinstance(geom.geometry, urdfpy.Box):
                color = (
                    Color(*material_dict.get(geom.material.name, (1, 1, 1, 1)))
                    if hasattr(geom, "material") and geom.material
                    else Color(1, 1, 1, 1)
                )
                res.append(
                    Box(
                        origin=origin_transform,
                        scale=Scale(*geom.geometry.size),
                        color=color,
                    )
                )
            elif isinstance(geom.geometry, urdfpy.Sphere):
                color = (
                    Color(*material_dict.get(geom.material.name, (1, 1, 1, 1)))
                    if hasattr(geom, "material") and geom.material
                    else Color(1, 1, 1, 1)
                )
                res.append(
                    Sphere(
                        origin=origin_transform,
                        radius=geom.geometry.radius,
                        color=color,
                    )
                )
            elif isinstance(geom.geometry, urdfpy.Cylinder):
                color = (
                    Color(*material_dict.get(geom.material.name, (1, 1, 1, 1)))
                    if hasattr(geom, "material") and geom.material
                    else Color(1, 1, 1, 1)
                )
                res.append(
                    Cylinder(
                        origin=origin_transform,
                        width=geom.geometry.radius,
                        height=geom.geometry.length,
                        color=color,
                    )
                )
            elif isinstance(geom.geometry, urdfpy.Mesh):
                if geom.geometry.filename is None:
                    raise ValueError("Mesh geometry must have a filename.")
                res.append(
                    FileMesh(
                        origin=origin_transform,
                        filename=self.parse_file_path(geom.geometry.filename),
                        scale=Scale(*(geom.geometry.scale or (1, 1, 1))),
                    )
                )
        return ShapeCollection(res, reference_frame=body)

    def parse_file_path(self, file_path: str) -> str:
        """
        Parses a file path which contains a ros package to a path in the local file system.

        :param file_path: The path to the URDF file.
        :return: The parsed and processed file path.
        """
        if "package://" in file_path:
            # Splits the file path at '//' to get the package  and the rest of the path
            package_split = file_path.split("//")
            # Splits the path after the // to get the package name and the rest of the path
            package_name = package_split[1].split("/")[0]
            try:
                from ament_index_python.packages import get_package_share_directory

                package_path = get_package_share_directory(package_name)
            except (ImportError, LookupError):
                if self.package_resolver:
                    if package_name in self.package_resolver:
                        package_path = self.package_resolver[package_name]
                    else:
                        raise ParsingError(
                            msg=f"Package '{package_name}' not found in package resolver and "
                            f"ROS is not installed."
                        )
                else:
                    raise ParsingError(
                        msg="No ROS install found while the URDF file contains references to "
                        "ROS packages."
                    )
            file_path = file_path.replace("package://" + package_name, package_path)
        if "file://" in file_path:
            file_path = file_path.replace("file://", "./")
        return file_path



import unittest

from semantic_digital_twin.world_description.degree_of_freedom import DegreeOfFreedom

from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.spatial_types.spatial_types import Vector3
from semantic_digital_twin.semantic_annotations.semantic_annotations import Handle, Door
from semantic_digital_twin.world import World
from semantic_digital_twin.world_description.connections import (
    FixedConnection,
    Connection6DoF,
    PrismaticConnection,
    RevoluteConnection,
)
from semantic_digital_twin.world_description.world_entity import Body
from semantic_digital_twin.world_description.world_modification import (
    WorldModelModificationBlock,
    AddKinematicStructureEntityModification,
    AddConnectionModification,
    AddDegreeOfFreedomModification,
    AddSemanticAnnotationModification,
    RemoveSemanticAnnotationModification,
)


class ConnectionModificationTestCase(unittest.TestCase):

    def test_single_modification(self):
        w = World()

        with w.modify_world():
            b1 = Body(name=PrefixedName("b1"))
            b2 = Body(name=PrefixedName("b2"))
            w.add_kinematic_structure_entity(b1)
            w.add_kinematic_structure_entity(b2)

            connection = FixedConnection(b1, b2)
            w.add_connection(connection)

    def test_ChangeDofHasHardwareInterface(self):
        w = World()

        with w.modify_world():
            b1 = Body(name=PrefixedName("b1"))
            b2 = Body(name=PrefixedName("b2"))
            w.add_kinematic_structure_entity(b1)
            w.add_kinematic_structure_entity(b2)

            dof = DegreeOfFreedom(name=PrefixedName("dofyboi"))
            w.add_degree_of_freedom(dof)
            connection = RevoluteConnection(
                b1, b2, axis=Vector3.from_iterable([0, 0, 1]), dof_name=dof.name
            )
            w.add_connection(connection)
        assert connection.dof.has_hardware_interface is False

        with w.modify_world():
            w.set_dofs_has_hardware_interface(connection.dofs, True)
        assert connection.dof.has_hardware_interface is True

    def test_many_modifications(self):
        w = World()

        with w.modify_world():
            b1 = Body(name=PrefixedName("b1"))
            b2 = Body(name=PrefixedName("b2"))
            b3 = Body(name=PrefixedName("b3"))
            w.add_kinematic_structure_entity(b1)
            w.add_kinematic_structure_entity(b2)
            w.add_kinematic_structure_entity(b3)
            w.add_connection(
                Connection6DoF.create_with_dofs(parent=b1, child=b2, world=w)
            )
            dof = DegreeOfFreedom(name=PrefixedName("dofyboi"))
            w.add_degree_of_freedom(dof)
            w.add_connection(
                PrismaticConnection(
                    parent=b2,
                    child=b3,
                    axis=Vector3.from_iterable([0, 0, 1]),
                    dof_name=dof.name,
                )
            )

        modifications = w._model_modification_blocks[-1]
        self.assertEqual(len(modifications.modifications), 13)

        add_body_modifications = [
            m
            for m in modifications.modifications
            if isinstance(m, AddKinematicStructureEntityModification)
        ]
        self.assertEqual(len(add_body_modifications), 3)

        add_dof_modifications = [
            m
            for m in modifications.modifications
            if isinstance(m, AddDegreeOfFreedomModification)
        ]
        self.assertEqual(len(add_dof_modifications), 8)

        add_connection_modifications = [
            m
            for m in modifications.modifications
            if isinstance(m, AddConnectionModification)
        ]
        self.assertEqual(len(add_connection_modifications), 2)

        # reconstruct this world
        w2 = World()

        # copy modifications
        modifications_copy = WorldModelModificationBlock.from_json(
            modifications.to_json()
        )
        modifications_copy.apply(w2)
        self.assertEqual(len(w2.bodies), 3)
        self.assertEqual(len(w2.connections), 2)

        with w.modify_world():
            w.remove_connection(w.connections[-1])
            w.remove_kinematic_structure_entity(
                w.get_kinematic_structure_entity_by_name("b3")
            )

        modifications = w._model_modification_blocks[-1]
        self.assertEqual(len(modifications.modifications), 3)

        modifications_copy = WorldModelModificationBlock.from_json(
            modifications.to_json()
        )
        modifications_copy.apply(w2)
        self.assertEqual(len(w2.bodies), 2)
        self.assertEqual(len(w2.connections), 1)

    def test_semantic_annotation_modifications(self):
        w = World()
        b1 = Body(name=PrefixedName("b1"))
        v1 = Handle(body=b1)
        v2 = Door(body=b1, handle=v1)

        add_v1 = AddSemanticAnnotationModification(v1)
        add_v2 = AddSemanticAnnotationModification(v2)

        self.assertNotIn(v1, w.semantic_annotations)
        self.assertNotIn(v2, w.semantic_annotations)

        with w.modify_world():
            add_v1.apply(w)
            add_v2.apply(w)

        self.assertIn(v1, w.semantic_annotations)
        self.assertIn(v2, w.semantic_annotations)

        rm_v1 = RemoveSemanticAnnotationModification(v1)
        rm_v2 = RemoveSemanticAnnotationModification(v2)
        with w.modify_world():
            rm_v1.apply(w)
            rm_v2.apply(w)

        self.assertNotIn(v1, w.semantic_annotations)
        self.assertNotIn(v2, w.semantic_annotations)


if __name__ == "__main__":
    unittest.main()



from dataclasses import dataclass, field
from io import BytesIO

import numpy as np
import trimesh
import trimesh.exchange.stl
from krrood.ormatic.dao import AlternativeMapping
from sqlalchemy import TypeDecorator, types
from typing_extensions import List
from typing_extensions import Optional

from ..datastructures.prefixed_name import PrefixedName
from ..spatial_types import RotationMatrix, Vector3, Point3, TransformationMatrix
from ..spatial_types.derivatives import DerivativeMap
from ..spatial_types.spatial_types import Quaternion
from ..spatial_types.symbol_manager import symbol_manager
from ..world import World
from ..world_description.connections import Connection
from ..world_description.degree_of_freedom import DegreeOfFreedom
from ..world_description.world_entity import (
    SemanticAnnotation,
    KinematicStructureEntity,
)
from ..world_description.world_state import WorldState


@dataclass
class WorldMapping(AlternativeMapping[World]):
    kinematic_structure_entities: List[KinematicStructureEntity]
    connections: List[Connection]
    semantic_annotations: List[SemanticAnnotation]
    degrees_of_freedom: List[DegreeOfFreedom]
    state: WorldState
    name: Optional[str] = field(default=None)

    @classmethod
    def create_instance(cls, obj: World):
        return cls(
            kinematic_structure_entities=obj.kinematic_structure_entities,
            connections=obj.connections,
            semantic_annotations=obj.semantic_annotations,
            degrees_of_freedom=list(obj.degrees_of_freedom),
            state=obj.state,
            name=obj.name,
        )

    def create_from_dao(self) -> World:
        result = World(name=self.name)

        with result.modify_world():
            for entity in self.kinematic_structure_entities:
                result.add_kinematic_structure_entity(entity)
            for dof in self.degrees_of_freedom:
                d = DegreeOfFreedom(
                    name=dof.name,
                    lower_limits=dof.lower_limits,
                    upper_limits=dof.upper_limits,
                )
                result.add_degree_of_freedom(d)
            for connection in self.connections:
                result.add_connection(connection)
            for semantic_annotation in self.semantic_annotations:
                result.add_semantic_annotation(semantic_annotation)
            result.delete_orphaned_dofs()
            result.state = self.state

        return result


@dataclass
class WorldStateMapping(AlternativeMapping[WorldState]):
    data: List[float]
    names: List[PrefixedName]

    @classmethod
    def create_instance(cls, obj: WorldState):
        return cls(
            data=obj.data.ravel().tolist(),
            names=obj._names,
        )

    def create_from_dao(self) -> WorldState:
        return WorldState(
            data=np.array(self.data, dtype=np.float64).reshape((4, len(self.names))),
            _names=self.names,
            _index={name: idx for idx, name in enumerate(self.names)},
        )


@dataclass
class Vector3Mapping(AlternativeMapping[Vector3]):
    x: float
    y: float
    z: float

    reference_frame: Optional[KinematicStructureEntity] = field(
        init=False, default=None
    )

    @classmethod
    def create_instance(cls, obj: Vector3):
        x, y, z, _ = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Vector3:
        return Vector3(
            x_init=self.x, y_init=self.y, z_init=self.z, reference_frame=None
        )


@dataclass
class Point3Mapping(AlternativeMapping[Point3]):
    x: float
    y: float
    z: float

    reference_frame: Optional[KinematicStructureEntity] = field(
        init=False, default=None
    )

    @classmethod
    def create_instance(cls, obj: Point3):
        x, y, z, _ = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Point3:
        return Point3(x_init=self.x, y_init=self.y, z_init=self.z, reference_frame=None)


@dataclass
class QuaternionMapping(AlternativeMapping[Quaternion]):
    x: float
    y: float
    z: float
    w: float

    reference_frame: Optional[KinematicStructureEntity] = field(
        init=False, default=None
    )

    @classmethod
    def create_instance(cls, obj: Quaternion):
        x, y, z, w = symbol_manager.evaluate_expr(obj).tolist()
        result = cls(x=x, y=y, z=z, w=w)
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> Quaternion:
        return Quaternion(
            x_init=self.x,
            y_init=self.y,
            z_init=self.z,
            w_init=self.w,
            reference_frame=None,
        )


@dataclass
class RotationMatrixMapping(AlternativeMapping[RotationMatrix]):
    rotation: Quaternion
    reference_frame: Optional[KinematicStructureEntity] = field(
        init=False, default=None
    )

    @classmethod
    def create_instance(cls, obj: RotationMatrix):
        result = cls(rotation=obj.to_quaternion())
        result.reference_frame = obj.reference_frame
        return result

    def create_from_dao(self) -> RotationMatrix:
        result = RotationMatrix.from_quaternion(self.rotation)
        result.reference_frame = None
        return result


@dataclass
class TransformationMatrixMapping(AlternativeMapping[TransformationMatrix]):
    position: Point3
    rotation: Quaternion
    reference_frame: Optional[KinematicStructureEntity] = field(
        init=False, default=None
    )
    child_frame: Optional[KinematicStructureEntity] = field(init=False, default=None)

    @classmethod
    def create_instance(cls, obj: TransformationMatrix):
        position = obj.to_position()
        rotation = obj.to_quaternion()
        result = cls(position=position, rotation=rotation)
        result.reference_frame = obj.reference_frame
        result.child_frame = obj.child_frame

        return result

    def create_from_dao(self) -> TransformationMatrix:
        return TransformationMatrix.from_point_rotation_matrix(
            point=self.position,
            rotation_matrix=RotationMatrix.from_quaternion(self.rotation),
            reference_frame=None,
            child_frame=self.child_frame,
        )


@dataclass
class DegreeOfFreedomMapping(AlternativeMapping[DegreeOfFreedom]):
    name: PrefixedName
    lower_limits: List[float]
    upper_limits: List[float]

    @classmethod
    def create_instance(cls, obj: DegreeOfFreedom):
        return cls(
            name=obj.name,
            lower_limits=obj.lower_limits.data,
            upper_limits=obj.upper_limits.data,
        )

    def create_from_dao(self) -> DegreeOfFreedom:
        lower_limits = DerivativeMap(data=self.lower_limits)
        upper_limits = DerivativeMap(data=self.upper_limits)
        return DegreeOfFreedom(
            name=self.name, lower_limits=lower_limits, upper_limits=upper_limits
        )


class TrimeshType(TypeDecorator):
    """
    Type that casts fields that are of type `type` to their class name on serialization and converts the name
    to the class itself through the globals on load.
    """

    impl = types.LargeBinary(4 * 1024 * 1024 * 1024 - 1)  # 4 GB max

    def process_bind_param(self, value: trimesh.Trimesh, dialect):
        # return binary version of trimesh
        return trimesh.exchange.stl.export_stl(value)

    def process_result_value(self, value: impl, dialect) -> Optional[trimesh.Trimesh]:
        if value is None:
            return None
        mesh = trimesh.Trimesh(**trimesh.exchange.stl.load_stl_binary(BytesIO(value)))
        return mesh



import re

from ...datastructures.prefixed_name import PrefixedName
from ...spatial_types.spatial_types import TransformationMatrix
from ...semantic_annotations.factories import (
    HandleFactory,
    ContainerFactory,
    Direction,
    DrawerFactory,
    DoorFactory,
    DresserFactory,
    SemanticPositionDescription,
    HorizontalSemanticDirection,
    VerticalSemanticDirection,
)
from ...world_description.geometry import Scale
from ...world_description.world_entity import Body


def drawer_factory_from_body(drawer: Body) -> DrawerFactory:
    """
    Create a DrawerFactory from a drawer body.
    This function assumes that the drawer body has a bounding box that can be used to determine its
    scale and that a handle can be created with a standard size.
    """
    handle_factory = HandleFactory(
        name=PrefixedName(drawer.name.name + "_handle", drawer.name.prefix),
        scale=Scale(0.05, 0.1, 0.02),
    )
    container_factory = ContainerFactory(
        name=PrefixedName(drawer.name.name + "_container", drawer.name.prefix),
        scale=drawer.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=drawer._world.root)
        )
        .bounding_boxes[0]
        .scale,
        direction=Direction.Z,
    )
    drawer_T_handle = TransformationMatrix.from_xyz_rpy(x=container_factory.scale.x / 2)
    drawer_factory = DrawerFactory(
        name=drawer.name,
        handle_factory=handle_factory,
        parent_T_handle=drawer_T_handle,
        container_factory=container_factory,
    )
    return drawer_factory


def door_factory_from_body(door: Body) -> DoorFactory:
    """
    Create a DoorFactory from a door body.
    This function assumes that the door body has a bounding box that can be used to determine its
    scale and that a handle can be created with a standard size.
    """
    handle_factory = HandleFactory(
        name=PrefixedName(door.name.name + "_handle", door.name.prefix),
        scale=Scale(0.05, 0.1, 0.02),
    )

    door_factory = DoorFactory(
        name=door.name,
        scale=door.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=door._world.root)
        )
        .bounding_boxes[0]
        .scale,
        handle_factory=handle_factory,
        semantic_position=SemanticPositionDescription(
            horizontal_direction_chain=[
                HorizontalSemanticDirection.RIGHT,
                HorizontalSemanticDirection.FULLY_CENTER,
            ],
            vertical_direction_chain=[VerticalSemanticDirection.FULLY_CENTER],
        ),
    )
    return door_factory


def dresser_factory_from_body(dresser: Body) -> DresserFactory:
    """
    Replace a dresser body with a DresserFactory.
    This function identifies drawers and doors in the dresser based on naming conventions
    and creates corresponding factories for them.
    It assumes that drawer bodies have names containing '_drawer_' and door bodies have names
    containing '_door_'.
    """
    drawer_pattern = re.compile(r"^.*_drawer_.*$")
    door_pattern = re.compile(r"^.*_door_.*$")
    drawer_factories = []
    drawer_transforms = []
    door_factories = []
    door_transforms = []
    for child in dresser._world.compute_child_kinematic_structure_entities(dresser):
        child: Body
        if bool(drawer_pattern.fullmatch(child.name.name)):
            drawer_transforms.append(child.parent_connection.origin_expression)
            drawer_factory = drawer_factory_from_body(child)
            drawer_factories.append(drawer_factory)
        elif bool(door_pattern.fullmatch(child.name.name)):
            door_transforms.append(child.parent_connection.origin_expression)
            door_factory = door_factory_from_body(child)
            door_factories.append(door_factory)

    dresser_container_factory = ContainerFactory(
        name=PrefixedName(dresser.name.name + "_container", dresser.name.prefix),
        scale=dresser.collision.as_bounding_box_collection_at_origin(
            TransformationMatrix(reference_frame=dresser._world.root)
        )
        .bounding_boxes[0]
        .scale,
        direction=Direction.X,
    )
    dresser_factory = DresserFactory(
        name=dresser.name,
        container_factory=dresser_container_factory,
        drawers_factories=drawer_factories,
        parent_T_drawers=drawer_transforms,
        door_factories=door_factories,
        door_transforms=door_transforms,
    )

    return dresser_factory



import os
import unittest

import trimesh.boolean

from semantic_digital_twin.adapters.mesh import STLParser
from semantic_digital_twin.world_description.geometry import Box
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.world_description.shape_collection import ShapeCollection
from semantic_digital_twin.world_description.world_entity import Body


class JSONTestCase(unittest.TestCase):

    def test_json_serialization(self):
        body = Body(name=PrefixedName("body"))
        collision = [
            Box(origin=TransformationMatrix.from_xyz_rpy(0, 1, 0, 0, 0, 1, body))
        ]
        body.collision = ShapeCollection(collision, reference_frame=body)

        json_data = body.to_json()
        body2 = Body.from_json(json_data)

        for c1 in body.collision:
            for c2 in body2.collision:
                self.assertEqual(c1, c2)

        self.assertEqual(body, body2)

    def test_json_serialization_with_mesh(self):
        body: Body = (
            STLParser(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "..",
                    "resources",
                    "stl",
                    "milk.stl",
                )
            )
            .parse()
            .root
        )

        json_data = body.to_json()
        body2 = Body.from_json(json_data)

        for c1 in body.collision:
            for c2 in body2.collision:
                assert (trimesh.boolean.difference([c1.mesh, c2.mesh])).is_empty



from __future__ import annotations

import logging
from abc import abstractmethod, ABC
from dataclasses import dataclass, field

from typing_extensions import (
    Iterable,
    Set,
    TYPE_CHECKING,
    Optional,
    Self,
    DefaultDict,
)

from ..spatial_types.derivatives import DerivativeMap
from ..spatial_types.spatial_types import (
    Vector3,
    Quaternion,
)
from ..world_description.connections import (
    ActiveConnection,
    OmniDrive,
    ActiveConnection1DOF,
)
from ..world_description.world_entity import (
    Body,
    RootedSemanticAnnotation,
    Connection,
    CollisionCheckingConfig,
)
from ..world_description.world_entity import (
    KinematicStructureEntity,
    Region,
)

if TYPE_CHECKING:
    from ..world import World


@dataclass
class SemanticRobotAnnotation(RootedSemanticAnnotation, ABC):
    """
    Represents a collection of connected robot bodies, starting from a root body, and ending in a unspecified collection
    of tip bodies.
    """

    _robot: AbstractRobot = field(default=None)
    """
    The robot this semantic annotation belongs to
    """

    def __post_init__(self):
        if self._world is not None:
            self._world.add_semantic_annotation(self, exists_ok=True)

    @abstractmethod
    def assign_to_robot(self, robot: AbstractRobot):
        """
        This method assigns the robot to the current semantic annotation, and then iterates through its own fields to call the
        appropriate methods to att them to the robot.

        :param robot: The robot to which this semantic annotation should be assigned.
        """
        ...


@dataclass
class KinematicChain(SemanticRobotAnnotation, ABC):
    """
    Abstract base class for kinematic chain in a robot, starting from a root body, and ending in a specific tip body.
    A kinematic chain can contain both a manipulator and sensors at the same time. There are no assumptions about the
    position of the manipulator or sensors in the kinematic chain
    """

    tip: Body = field(default=None)
    """
    The tip body of the kinematic chain, which is the last body in the chain.
    """

    manipulator: Optional[Manipulator] = None
    """
    The manipulator of the kinematic chain, if it exists. This is usually a gripper or similar device.
    """

    sensors: Set[Sensor] = field(default_factory=set)
    """
    A collection of sensors in the kinematic chain, such as cameras or other sensors.
    """

    @property
    def bodies(self) -> Iterable[Body]:
        """
        Returns itself as a kinematic chain of bodies.
        """
        return [
            entity
            for entity in self._world.compute_chain_of_kinematic_structure_entities(
                self.root, self.tip
            )
            if isinstance(entity, Body)
        ]

    @property
    def kinematic_structure_entities(self) -> Iterable[KinematicStructureEntity]:
        """
        Returns itself as a kinematic chain of KinematicStructureEntity.
        """
        return self._world.compute_chain_of_kinematic_structure_entities(
            self.root, self.tip
        )

    @property
    def regions(self) -> Iterable[Region]:
        """
        Returns itself as a kinematic chain of KinematicStructureEntity.
        """
        return [
            entity
            for entity in self._world.compute_chain_of_kinematic_structure_entities(
                self.root, self.tip
            )
            if isinstance(entity, Region)
        ]

    @property
    def connections(self) -> Iterable[Connection]:
        """
        Returns the connections of the kinematic chain.
        This is a list of connections between the bodies in the kinematic chain
        """
        return self._world.compute_chain_of_connections(self.root, self.tip)

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the kinematic chain to the given robot. This method ensures that the kinematic chain is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(
                f"Kinematic chain {self.name} is already part of another robot: {self._robot.name}."
            )
        if self._robot is not None:
            return
        self._robot = robot
        if self.manipulator is not None:
            robot.add_manipulator(self.manipulator)
        for sensor in self.sensors:
            robot.add_sensor(sensor)

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Arm(KinematicChain):
    """
    Represents an arm of a robot, which is a kinematic chain with a specific tip body.
    An arm has a manipulators and potentially sensors.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Manipulator(SemanticRobotAnnotation, ABC):
    """
    Abstract base class of robot manipulators. Always has a tool frame.
    """

    tool_frame: Body = field(default=None)

    front_facing_orientation: Quaternion = field(default=None)
    """
    The orientation of the manipulator's tool frame, which is usually the front-facing orientation.
    """

    front_facing_axis: Vector3 = field(default=None)
    """
    The axis of the manipulator's tool frame that is facing forward.
    """

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the manipulator to the given robot. This method ensures that the manipulator is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(
                f"Manipulator {self.name} is already part of another robot: {self._robot.name}."
            )
        if self._robot is not None:
            return
        self._robot = robot

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tool_frame))


@dataclass
class Finger(KinematicChain):
    """
    A finger is a kinematic chain, since it should have an unambiguous tip body, and may contain sensors.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class ParallelGripper(Manipulator):
    """
    Represents a gripper of a robot. Contains a collection of fingers and a thumb. The thumb is a specific finger
    that always needs to touch an object when grasping it, ensuring a stable grasp.
    """

    finger: Finger = field(default=None)
    thumb: Finger = field(default=None)

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the parallel gripper to the given robot and calls the appropriate methods for the its finger and thumb.
         This method ensures that the parallel gripper is only assigned to one robot at a time, and raises an error if
         it is already assigned to another
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(
                f"ParallelGripper {self.name} is already part of another robot: {self._robot.name}."
            )
        if self._robot is not None:
            return
        self._robot = robot

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tool_frame))


@dataclass
class Sensor(SemanticRobotAnnotation, ABC):
    """
    Abstract base class for any kind of sensor in a robot.
    """

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the sensor to the given robot. This method ensures that the sensor is only assigned
        to one robot at a time, and raises an error if it is already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(
                f"Sensor {self.name} is already part of another robot: {self._robot.name}."
            )
        if self._robot is not None:
            return
        self._robot = robot


@dataclass
class FieldOfView:
    """
    Represents the field of view of a camera sensor, defined by the vertical and horizontal angles of the camera's view.
    """

    vertical_angle: float
    horizontal_angle: float


@dataclass
class Camera(Sensor):
    """
    Represents a camera sensor in a robot.
    """

    forward_facing_axis: Vector3 = field(default=None)
    field_of_view: FieldOfView = field(default=None)
    minimal_height: float = 0.0
    maximal_height: float = 1.0

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root))


@dataclass
class Neck(KinematicChain):
    """
    Represents a special kinematic chain that connects the head of a robot with a collection of sensors, such as cameras
    and which does not have a manipulator.
    """

    pitch_body: Optional[Body] = None
    """
    The body that allows pitch movement in the neck, if it exists.
    """
    yaw_body: Optional[Body] = None
    """
    The body that allows yaw movement in the neck, if it exists.
    """

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class Torso(KinematicChain):
    """
    A Torso is a kinematic chain connecting the base of the robot with a collection of other kinematic chains.
    """

    def assign_to_robot(self, robot: AbstractRobot):
        """
        Assigns the torso to the given robot and calls the appropriate method for each of its attached kinematic chains.
         This method ensures that the torso is only assigned to one robot at a time, and raises an error if it is
         already assigned to another robot.
        """
        if self._robot is not None and self._robot != robot:
            raise ValueError(
                f"Torso {self.name} is already part of another robot: {self._robot.name}."
            )
        if self._robot is not None:
            return
        self._robot = robot

    def __hash__(self):
        """
        Returns the hash of the kinematic chain, which is based on the root and tip bodies.
        This allows for proper comparison and storage in sets or dictionaries.
        """
        return hash((self.name, self.root, self.tip))


@dataclass
class AbstractRobot(RootedSemanticAnnotation, ABC):
    """
    Specification of an abstract robot. A robot consists of:
    - a root body, which is the base of the robot
    - an optional torso, which is a kinematic chain (usually without a manipulator) connecting the base with a collection
        of other kinematic chains
    - an optional collection of manipulator chains, each containing a manipulator, such as a gripper
    - an optional collection of sensor chains, each containing a sensor, such as a camera
    => If a kinematic chain contains both a manipulator and a sensor, it will be part of both collections
    """

    torso: Optional[Torso] = None
    """
    The torso of the robot, which is a kinematic chain connecting the base with a collection of other kinematic chains.
    """

    manipulators: Set[Manipulator] = field(default_factory=set)
    """
    A collection of manipulators in the robot, such as grippers.
    """

    sensors: Set[Sensor] = field(default_factory=set)
    """
    A collection of sensors in the robot, such as cameras.
    """

    manipulator_chains: Set[KinematicChain] = field(default_factory=set)
    """
    A collection of all kinematic chains containing a manipulator, such as a gripper.
    """

    sensor_chains: Set[KinematicChain] = field(default_factory=set)
    """
    A collection of all kinematic chains containing a sensor, such as a camera.
    """

    default_collision_config: CollisionCheckingConfig = field(
        kw_only=True,
        default_factory=lambda: CollisionCheckingConfig(buffer_zone_distance=0.05),
    )

    @abstractmethod
    def load_srdf(self):
        """
        Loads the SRDF file for the robot, if it exists. This method is expected to be implemented in subclasses.
        """
        ...

    @property
    def controlled_connections(self) -> Set[ActiveConnection]:
        """
        A subset of the robot's connections that are controlled by a controller.
        """
        return self._world.controlled_connections & set(self.connections)

    @classmethod
    @abstractmethod
    def from_world(cls, world: World) -> Self:
        """
        Creates a robot semantic annotation from the given world.
        This method constructs the robot semantic annotation by identifying and organizing the various semantic components of the robot,
        such as manipulators, sensors, and kinematic chains. It is expected to be implemented in subclasses.

        :param world: The world from which to create the robot semantic annotation.

        :return: A robot semantic annotation.
        """
        raise NotImplementedError("This method should be implemented in subclasses.")

    @property
    def drive(self) -> Optional[OmniDrive]:
        """
        The connection which the robot uses for driving.
        """
        try:
            parent_connection = self.root.parent_connection
            if isinstance(parent_connection, OmniDrive):
                return parent_connection
        except AttributeError:
            pass

    def tighten_dof_velocity_limits_of_1dof_connections(
        self,
        new_limits: DefaultDict[ActiveConnection1DOF, float],
    ):
        """
        Convenience method for tightening the velocity limits of all one degree-of-freedom (1DOF)
        active connections in the system.

        The method iterates through all connections of type `ActiveConnection1DOF`
        and configures their velocity limits by overwriting the existing
        lower and upper limit values with the provided ones.

        :param new_limits: A dictionary linking 1DOF connections to their corresponding
            new velocity limits. The keys are of type `ActiveConnection1DOF`, and the
            values represent the new velocity limits specific to each connection.
        """
        for connection in self._world.get_connections_by_type(ActiveConnection1DOF):
            connection.raw_dof._overwrite_dof_limits(
                new_lower_limits=DerivativeMap(
                    [None, -new_limits[connection], None, None]
                ),
                new_upper_limits=DerivativeMap(
                    [None, new_limits[connection], None, None]
                ),
            )

    def add_manipulator(self, manipulator: Manipulator):
        """
        Adds a manipulator to the robot's collection of manipulators.
        """
        self.manipulators.add(manipulator)
        self._semantic_annotations.add(manipulator)
        manipulator.assign_to_robot(self)

    def add_sensor(self, sensor: Sensor):
        """
        Adds a sensor to the robot's collection of sensors.
        """
        self.sensors.add(sensor)
        self._semantic_annotations.add(sensor)
        sensor.assign_to_robot(self)

    def add_torso(self, torso: Torso):
        """
        Adds a torso to the robot's collection of kinematic chains.
        """
        if self.torso is not None:
            raise ValueError(
                f"Robot {self.name} already has a torso: {self.torso.name}."
            )
        self.torso = torso
        self._semantic_annotations.add(torso)
        torso.assign_to_robot(self)

    def add_kinematic_chain(self, kinematic_chain: KinematicChain):
        """
        Adds a kinematic chain to the robot's collection of kinematic chains.
        This can be either a manipulator chain or a sensor chain.
        """
        if kinematic_chain.manipulator is None and not kinematic_chain.sensors:
            logging.warning(
                f"Kinematic chain {kinematic_chain.name} has no manipulator or sensors, so it was skipped. Did you mean to add it to the torso?"
            )
            return
        if kinematic_chain.manipulator is not None:
            self.manipulator_chains.add(kinematic_chain)
        if kinematic_chain.sensors:
            self.sensor_chains.add(kinematic_chain)
        self._semantic_annotations.add(kinematic_chain)
        kinematic_chain.assign_to_robot(self)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np
from typing_extensions import TYPE_CHECKING, Callable, Dict

from ..datastructures.prefixed_name import PrefixedName

if TYPE_CHECKING:
    from ..world import World

logger = logging.getLogger(__name__)


@dataclass
class Callback(ABC):
    """
    Callback is an abstract base class (ABC)
    reacting to changes in the associated `world`.
    It provides a flexible  mechanism for subclasses to implement custom behaviors to be triggered
    whenever a change occurs.

    The primary purpose of this class is to encapsulate logic that needs to be
    executed as a response to certain events or changes within the `world` object.
    """

    world: World
    """
    The world this callback is listening on.
    """

    _is_paused = False
    """
    Flag that indicates if the callback is paused.
    """

    def notify(self):
        """
        Notify the callback of a change in the world.
        """
        if self._is_paused:
            pass
        else:
            self._notify()

    @abstractmethod
    def _notify(self):
        """
        Notify the callback of a change in the world.
        Override this method to implement custom behaviors.
        """
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        """
        Stop the callback.
        """
        raise NotImplementedError

    def pause(self):
        """
        Pause the callback such that notify does not trigger anymore.
        """
        self._is_paused = True

    def resume(self):
        """
        Resume the callback such that notify does trigger again.
        """
        self._is_paused = False


@dataclass
class StateChangeCallback(Callback, ABC):
    """
    Callback for handling state changes.
    """

    previous_world_state_data: np.ndarray = field(init=False)
    """
    The previous world state data used to check if something changed.
    """

    def __post_init__(self):
        self.world.state_change_callbacks.append(self)
        self.update_previous_world_state()

    def stop(self):
        try:
            self.world.state_change_callbacks.remove(self)
        except ValueError:
            pass

    def update_previous_world_state(self):
        """
        Update the previous world state to reflect the current world positions.
        """
        self.previous_world_state_data = np.copy(self.world.state.positions)

    def compute_state_changes(self) -> Dict[PrefixedName, float]:
        changes = {
            name: current_state
            for name, previous_state, current_state in zip(
                self.world.state.keys(),
                self.previous_world_state_data,
                self.world.state.positions,
            )
            if not np.allclose(previous_state, current_state)
        }
        return changes


@dataclass
class ModelChangeCallback(Callback, ABC):
    """
    Callback for handling model changes.
    """

    def __post_init__(self):
        self.world.model_change_callbacks.append(self)

    def stop(self):
        try:
            self.world.model_change_callbacks.remove(self)
        except ValueError:
            pass



import json
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import cached_property
from typing import ClassVar, Optional, Type, List

import numpy as np
import rclpy  # type: ignore
import std_msgs.msg
from krrood.ormatic.dao import to_dao
from random_events.utils import SubclassJSONSerializer
from rclpy.node import Node as RosNode
from rclpy.publisher import Publisher
from rclpy.subscription import Subscription
from sqlalchemy import select
from sqlalchemy.orm import Session

from .messages import MetaData, WorldStateUpdate, Message, ModificationBlock, LoadModel
from ...callbacks.callback import Callback, StateChangeCallback, ModelChangeCallback
from ...orm.ormatic_interface import *
from ...world import World


@dataclass
class Synchronizer(ABC):
    """
    Abstract Synchronizer class to manage world synchronizations between processes running semantic digital twin.
    It manages publishers and subscribers, ensuring proper cleanup after use.
    The communication is JSON string based.
    """

    node: RosNode
    """
    The rclpy node used to create the publishers and subscribers.
    """

    world: World
    """
    The world to synchronize.
    """

    topic_name: Optional[str] = None
    """
    The topic name of the publisher and subscriber.
    """

    publisher: Optional[Publisher] = field(init=False, default=None)
    """
    The publisher used to publish the world state.
    """

    subscriber: Optional[Subscription] = field(default=None, init=False)
    """
    The subscriber to the world state.
    """

    message_type: ClassVar[Optional[Type[SubclassJSONSerializer]]] = None
    """
    The type of the message that is sent and received.
    """

    def __post_init__(self):
        self.publisher = self.node.create_publisher(
            std_msgs.msg.String, topic=self.topic_name, qos_profile=10
        )
        self.subscriber = self.node.create_subscription(
            std_msgs.msg.String,
            topic=self.topic_name,
            callback=self.subscription_callback,
            qos_profile=10,
        )

    @cached_property
    def meta_data(self) -> MetaData:
        """
        The metadata of the synchronizer which can be used to compare origins of messages.
        """
        return MetaData(
            node_name=self.node.get_name(), process_id=os.getpid(), object_id=id(self)
        )

    def subscription_callback(self, msg: std_msgs.msg.String):
        """
        Wrap the origin subscription callback by self-skipping and disabling the next world callback.
        """
        msg = self.message_type.from_json(json.loads(msg.data))
        if msg.meta_data == self.meta_data:
            return
        self._skip_next_world_callback = True
        self._subscription_callback(msg)

    @abstractmethod
    def _subscription_callback(self, msg: message_type):
        """
        Callback function called when receiving new messages from other publishers.
        """
        raise NotImplementedError

    def publish(self, msg: Message):
        self.publisher.publish(std_msgs.msg.String(data=json.dumps(msg.to_json())))

    def close(self):
        """
        Clean up publishers and subscribers.
        """

        # Destroy subscribers
        if self.subscriber is not None:
            self.node.destroy_subscription(self.subscriber)
            self.subscriber = None

        # Destroy publishers
        if self.publisher is not None:
            self.node.destroy_publisher(self.publisher)
            self.publisher = None


@dataclass
class SynchronizerOnCallback(Synchronizer, Callback, ABC):
    """
    Synchronizer that does something on callbacks by the world.
    Additionally, ensures that the callback is cleaned up on close.
    """

    _skip_next_world_callback: bool = False
    """
    Flag to indicate if the next world callback should be skipped.
    
    An incoming message from some other world might trigger a change in this world that produces a notify callback that 
    will try to send a message. 
    If the callback is triggered by a message, this synchronizer should not republish the change.
    """

    missed_messages: List = field(default_factory=list, init=False, repr=False)
    """
    The messages that the callback did not trigger due to being paused.
    """

    def _notify(self):
        """
        Wrapper method around world_callback that checks if this time the callback should be triggered.
        """
        if self._skip_next_world_callback:
            self._skip_next_world_callback = False
        else:
            self.world_callback()

    def _subscription_callback(self, msg):
        if self._is_paused:
            self.missed_messages.append(msg)
        else:
            self.apply_message(msg)

    @abstractmethod
    def apply_message(self, msg):
        """
        Apply the received message to the world.
        """
        raise NotImplementedError

    @abstractmethod
    def world_callback(self):
        """
        Called when the world notifies and update that is not caused by this synchronizer.
        """
        raise NotImplementedError

    def apply_missed_messages(self):
        """
        Applies the missed messages to the world.
        """
        self._skip_next_world_callback = True
        with self.world.modify_world():
            for msg in self.missed_messages:
                self.apply_message(msg)

        self.missed_messages = []


@dataclass
class StateSynchronizer(StateChangeCallback, SynchronizerOnCallback):
    """
    Synchronizes the state (values of free variables) of the semantic digital twin with the associated ROS topic.
    """

    message_type: ClassVar[Optional[Type[SubclassJSONSerializer]]] = WorldStateUpdate

    topic_name: str = "/semantic_digital_twin/world_state"

    def __post_init__(self):
        super().__post_init__()
        SynchronizerOnCallback.__post_init__(self)

    def apply_message(self, msg: WorldStateUpdate):
        """
        Update the world state with the provided message.

        :param msg: The message containing the new state information.
        """
        # Parse incoming states: WorldState has 'states' only
        indices = [self.world.state._index[n] for n in msg.prefixed_names]

        if indices:
            self.world.state.data[0, indices] = np.asarray(msg.states, dtype=float)
            self.update_previous_world_state()
            self.world.notify_state_change()

    def world_callback(self):
        """
        Publish the current world state to the ROS topic.
        """
        changes = self.compute_state_changes()

        if not changes:
            return

        msg = WorldStateUpdate(
            prefixed_names=list(changes.keys()),
            states=list(changes.values()),
            meta_data=self.meta_data,
        )
        self.update_previous_world_state()
        self.publish(msg)


@dataclass
class ModelSynchronizer(
    ModelChangeCallback,
    SynchronizerOnCallback,
):
    """
    Synchronizes the model (addition/removal of bodies/DOFs/connections) with the associated ROS topic.
    """

    message_type: ClassVar[Type[SubclassJSONSerializer]] = ModificationBlock
    topic_name: str = "/semantic_digital_twin/world_model"

    def __post_init__(self):
        super().__post_init__()
        SynchronizerOnCallback.__post_init__(self)

    def apply_message(self, msg: ModificationBlock):
        msg.modifications.apply(self.world)

    def world_callback(self):
        msg = ModificationBlock(
            meta_data=self.meta_data,
            modifications=self.world._model_modification_blocks[-1],
        )
        self.publish(msg)


@dataclass
class ModelReloadSynchronizer(Synchronizer):
    """
    Synchronizes the model reloading process across different systems using ROS messaging.
    The database must be the same across the different processes, otherwise the synchronizer will fail.

    Use this when you did changes to the model that cannot be communicated via the ModelSynchronizer and hence need
    to force all processes to load your world model. Note that this may take a couple of seconds.
    """

    message_type: ClassVar[Type[SubclassJSONSerializer]] = LoadModel

    session: Session = None
    """
    The session used to perform persistence interaction. 
    """

    topic_name: str = "/semantic_digital_twin/reload_model"

    def __post_init__(self):
        super().__post_init__()
        assert self.session is not None

    def publish_reload_model(self):
        """
        Save the current world model to the database and publish the primary key to the ROS topic such that other
        processes can subscribe to the model changes and update their worlds.
        """
        dao: WorldMappingDAO = to_dao(self.world)
        self.session.add(dao)
        self.session.commit()
        message = LoadModel(primary_key=dao.database_id, meta_data=self.meta_data)
        self.publish(message)

    def _subscription_callback(self, msg: LoadModel):
        """
        Update the world with the new model by fetching it from the database.

        :param msg: The message containing the primary key of the model to be fetched.
        """
        query = select(WorldMappingDAO).where(
            WorldMappingDAO.database_id == msg.primary_key
        )
        new_world = self.session.scalars(query).one().from_dao()
        self._replace_world(new_world)
        self.world._notify_model_change()

    def _replace_world(self, new_world: World):
        """
        Replaces the current world with a new one, updating all relevant attributes.
        This method modifies the existing world state, kinematic structure, degrees
        of freedom, and semantic annotation based on the `new_world` provided.

        If you encounter any issues with references to dead objects, it is most likely due to this method not doing
        everything needed.

        :param new_world: The new world instance to replace the current world.
        """
        with self.world.modify_world():
            self.world.clear()
            self.world.merge_world(new_world)



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Optional, List

import coacd
import numpy as np
import trimesh

from .pipeline import Step
from ..world import World
from ..world_description.geometry import TriangleMesh, Shape, Mesh
from ..world_description.world_entity import Body


class ApproximationMode(StrEnum):
    """
    Approximation shape type
    """

    BOX = "box"
    CONVEX_HULL = "ch"


class PreprocessingMode(StrEnum):
    """
    Manifold preprocessing mode
    """

    AUTO = "auto"
    """
    Automatically chose based on the geometry.
    """

    ON = "on"
    """
    Force turn on the pre-processing
    """

    OFF = "off"
    """
    Force turn off the pre-processing
    """


@dataclass
class MeshDecomposer(Step, ABC):
    """
    MeshDecomposer is an abstract base class for decomposing complex 3D meshes into simpler convex components.
    It provides methods to apply the decomposition to meshes, shapes, bodies, and entire worlds.
    Subclasses should implement the `apply_to_mesh` method to define the specific decomposition algorithm.
    3D meshes are represented using the `trimesh` library, and the decomposed parts are returned as a list of `TriangleMesh` objects.
    """

    @abstractmethod
    def apply_to_mesh(self, mesh: Mesh) -> List[TriangleMesh]:
        """
        Apply the mesh decomposition to a given mesh.
        Returns a list of TriangleMesh objects representing the decomposed convex parts.
        """
        ...

    def apply_to_shape(self, shape: Shape) -> List[Shape]:
        """
        Apply the mesh decomposition to a given shape.
        If the shape is a Mesh, it will be decomposed into multiple TriangleMesh objects.
        Otherwise, the shape will be returned as is in a list.
        """
        new_geometry = []
        if isinstance(shape, Mesh):
            self.apply_to_mesh(shape)
        else:
            new_geometry.append(shape)

        return new_geometry

    def apply_to_body(self, body: Body) -> Body:
        """
        Apply the mesh decomposition to all shapes in a given body.
        The body's collision shapes will be replaced with the decomposed shapes.
        Returns the modified body.
        """
        new_geometry = []
        for shape in body.visual:
            decomposed_shapes = self.apply_to_shape(shape)
            new_geometry.extend(decomposed_shapes)

        body.collision = new_geometry
        return body

    def _apply(self, world: World) -> World:
        """
        Apply the mesh decomposition to all bodies in a given world.
        Each body's collision shapes will be replaced with the decomposed shapes.
        Returns the modified world.
        """
        for body in world.bodies:
            self.apply_to_body(body)

        return world


@dataclass
class COACDMeshDecomposer(MeshDecomposer):
    """
    COACDMeshDecomposer is a class for decomposing complex 3D meshes into simpler convex components
    using the COACD (Convex Optimization for Approximate Convex Decomposition) algorithm. It is
    designed to preprocess, analyze, and process 3D meshes with a focus on efficiency and scalability
    in fields such as robotics, gaming, and simulation.

    Check https://github.com/SarahWeiii/CoACD for further details.
    """

    threshold: float = 0.05
    """
    Concavity threshold for terminating the decomposition (0.01 - 1)
    """

    max_convex_hull: Optional[int] = None
    """
    Maximum number of convex hulls in the result. 
    Works only when merge is enabled (may introduce convex hull with a concavity larger than the threshold)
    """

    preprocess_mode: PreprocessingMode = PreprocessingMode.AUTO
    """
    Manifold preprocessing mode.
    """

    preprocess_resolution: int = 50
    """
    Resolution for manifold preprocess (20~100)
    """

    resolution: int = 2000
    """
    Sampling resolution for Hausdorff distance calculation (1 000 - 10 000)
    """

    search_nodes: int = 20
    """
    Max number of child nodes in the monte carlo tree search (10 - 40).
    """

    search_iterations: int = 150
    """
    Number of search iterations in the monte carlo tree search (60 - 2000).
    """

    search_depth: int = 3
    """
    Maximum search depth in the monte carlo tree search (2 - 7).
    """

    pca: bool = False
    """
    Enable PCA pre-processing
    """

    merge: bool = True
    """
    Enable merge postprocessing.
    """

    max_convex_hull_vertices: Optional[int] = None
    """
    Maximum vertex value for each convex hull, only when decimate is enabled.
    """

    extrude_margin: Optional[float] = None
    """
    Extrude margin, only when extrude is enabled
    """

    approximation_mode: ApproximationMode = ApproximationMode.BOX
    """
    Approximation mode to use.
    """

    seed: int = field(default_factory=lambda: np.random.randint(2**32))
    """
    Random seed used for sampling.
    """

    def apply_to_mesh(self, mesh: Mesh) -> List[TriangleMesh]:
        """
        Apply the COACD mesh decomposition to a given mesh.
        Returns a list of TriangleMesh objects representing the decomposed convex parts.
        """
        new_geometry = []

        trimesh_mesh = mesh.mesh
        if mesh.scale.x == mesh.scale.y == mesh.scale.z:
            trimesh_mesh.apply_scale(mesh.scale.x)
        else:
            logging.warning("Ambiguous scale for mesh, using uniform scale only.")

        coacd_mesh = coacd.Mesh(trimesh_mesh.vertices, trimesh_mesh.faces)
        if self.max_convex_hull is not None:
            max_convex_hull = self.max_convex_hull
        else:
            max_convex_hull = -1
        parts = coacd.run_coacd(
            mesh=coacd_mesh,
            apx_mode=str(self.approximation_mode),
            threshold=self.threshold,
            max_convex_hull=max_convex_hull,
            preprocess_mode=str(self.preprocess_mode),
            resolution=self.resolution,
            mcts_nodes=self.search_nodes,
            mcts_iterations=self.search_iterations,
            mcts_max_depth=self.search_depth,
            pca=self.pca,
            merge=self.merge,
            decimate=self.max_convex_hull_vertices is not None,
            max_ch_vertex=self.max_convex_hull_vertices or 256,
            extrude=self.extrude_margin is not None,
            extrude_margin=self.extrude_margin or 0.01,
            seed=self.seed,
        )

        for vs, fs in parts:
            new_geometry.append(
                TriangleMesh(mesh=trimesh.Trimesh(vs, fs), origin=mesh.origin)
            )

        return new_geometry



import unittest

from random_events.variable import Continuous
from sortedcontainers import SortedSet

from semantic_digital_twin.datastructures.variables import SpatialVariables


class TestSpatialVariables(unittest.TestCase):
    def test_enum_members_exist(self):

        self.assertTrue(hasattr(SpatialVariables, "x"))
        self.assertTrue(hasattr(SpatialVariables, "y"))
        self.assertTrue(hasattr(SpatialVariables, "z"))

        self.assertEqual(SpatialVariables.x.value, Continuous("x"))
        self.assertEqual(SpatialVariables.y.value, Continuous("y"))
        self.assertEqual(SpatialVariables.z.value, Continuous("z"))

        # Distinct members and distinct underlying values
        self.assertNotEqual(SpatialVariables.x, SpatialVariables.y)
        self.assertNotEqual(SpatialVariables.x, SpatialVariables.z)
        self.assertNotEqual(SpatialVariables.y, SpatialVariables.z)

        values = [
            SpatialVariables.x.value,
            SpatialVariables.y.value,
            SpatialVariables.z.value,
        ]
        self.assertEqual(
            len(set(values)), 3, "Underlying values for x, y, z should be distinct"
        )

    def test_xy_contains_correct_members(self):
        xy = SpatialVariables.xy
        expected = {SpatialVariables.x.value, SpatialVariables.y.value}
        self.assertEqual(set(xy), expected, "xy should contain exactly x and y values")
        if SortedSet is not None:
            self.assertIsInstance(xy, SortedSet)

    def test_xz_contains_correct_members(self):
        xz = SpatialVariables.xz
        expected = {SpatialVariables.x.value, SpatialVariables.z.value}
        self.assertEqual(set(xz), expected, "xz should contain exactly x and z values")
        if SortedSet is not None:
            self.assertIsInstance(xz, SortedSet)

    def test_yz_contains_correct_members(self):
        yz = SpatialVariables.yz
        expected = {SpatialVariables.y.value, SpatialVariables.z.value}
        self.assertEqual(set(yz), expected, "yz should contain exactly y and z values")
        if SortedSet is not None:
            self.assertIsInstance(yz, SortedSet)


if __name__ == "__main__":
    unittest.main()



import itertools

import pytest

from semantic_digital_twin.collision_checking.collision_detector import CollisionCheck
from semantic_digital_twin.collision_checking.trimesh_collision_detector import (
    TrimeshCollisionDetector,
)
from semantic_digital_twin.spatial_types import TransformationMatrix
from semantic_digital_twin.testing import world_setup_simple


def test_simple_collision(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    tcd = TrimeshCollisionDetector(world)
    collision = tcd.check_collision_between_bodies(body1, body2)
    assert collision
    assert collision.body_a == body1
    assert collision.body_b == body2


def test_no_collision(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    body1.parent_connection.origin = TransformationMatrix.from_xyz_rpy(1, 1, 1)
    tcd = TrimeshCollisionDetector(world)
    collision = tcd.check_collision_between_bodies(body1, body2)
    assert not collision


@pytest.mark.skip(reason="Not my test not my problem.")
def test_collision_matrix(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    tcd = TrimeshCollisionDetector(world)
    collisions = tcd.check_collisions(
        {
            CollisionCheck(body1, body2, 0.0, world),
            CollisionCheck(body3, body4, 0.0, world),
        }
    )
    assert len(collisions) == 2
    pairs = {(c.body_a, c.body_b) for c in collisions}
    assert (body1, body2) in pairs
    assert (body3, body4) in pairs


def test_all_collisions(world_setup_simple):
    world, body1, body2, body3, body4 = world_setup_simple
    tcd = TrimeshCollisionDetector(world)
    body4.parent_connection.origin = TransformationMatrix.from_xyz_rpy(10, 10, 10)
    body3.parent_connection.origin = TransformationMatrix.from_xyz_rpy(-10, -10, 10)

    collisions = tcd.check_collisions(
        [
            CollisionCheck(a, b, _world=world, distance=0.0001)
            for a, b in itertools.combinations(world.bodies_with_enabled_collision, 2)
        ]
    )
    assert len(collisions) == 1
    assert collisions[0].body_a == body1
    assert collisions[0].body_b == body2



import logging
import os
from dataclasses import dataclass
from typing_extensions import Optional

import numpy

try:
    from multiverse_parser import (
        InertiaSource,
        UsdImporter,
        MjcfImporter,
        UrdfImporter,
        BodyBuilder,
        JointBuilder,
        JointType,
    )
    from pxr import UsdUrdf
except ImportError as e:
    logging.info(e)
    InertiaSource = None
    UsdImporter = None
    MjcfImporter = None
    UrdfImporter = None
    BodyBuilder = None
    JointBuilder = None
    JointType = None
    UsdUrdf = None

from ..world_description.connections import (
    RevoluteConnection,
    PrismaticConnection,
    FixedConnection,
)
from ..world_description.degree_of_freedom import DegreeOfFreedom
from ..datastructures.prefixed_name import PrefixedName
from ..spatial_types import spatial_types as cas
from ..spatial_types.derivatives import DerivativeMap
from ..world import World, Body, Connection


@dataclass
class MultiParser:
    """
    Class to parse any scene description files to worlds.
    """

    file_path: str
    """
    The file path of the scene.
    """

    prefix: Optional[str] = None
    """
    The prefix for every name used in this world.
    """

    def __post_init__(self):
        if self.prefix is None:
            self.prefix = os.path.basename(self.file_path).split(".")[0]

    def parse(self) -> World:
        fixed_base = True
        root_name = None
        with_physics = True
        with_visual = True
        with_collision = True
        inertia_source = InertiaSource.FROM_SRC
        default_rgba = numpy.array([0.9, 0.9, 0.9, 1.0])

        file_ext = os.path.splitext(self.file_path)[1]
        if file_ext in [".usd", ".usda", ".usdc"]:
            add_xform_for_each_geom = True
            factory = UsdImporter(
                file_path=self.file_path,
                fixed_base=fixed_base,
                root_name=root_name,
                with_physics=with_physics,
                with_visual=with_visual,
                with_collision=with_collision,
                inertia_source=inertia_source,
                default_rgba=default_rgba,
                add_xform_for_each_geom=add_xform_for_each_geom,
            )
        elif file_ext == ".urdf":
            factory = UrdfImporter(
                file_path=self.file_path,
                fixed_base=fixed_base,
                root_name=root_name,
                with_physics=with_physics,
                with_visual=with_visual,
                with_collision=with_collision,
                inertia_source=inertia_source,
                default_rgba=default_rgba,
            )
        elif file_ext == ".xml":
            if root_name is None:
                root_name = "world"
            factory = MjcfImporter(
                file_path=self.file_path,
                fixed_base=fixed_base,
                root_name=root_name,
                with_physics=with_physics,
                with_visual=with_visual,
                with_collision=with_collision,
                inertia_source=inertia_source,
                default_rgba=default_rgba,
            )
        else:
            raise NotImplementedError(
                f"Importing from {file_ext} is not supported yet."
            )

        factory.import_model()
        bodies = [
            self.parse_body(body_builder)
            for body_builder in factory.world_builder.body_builders
        ]
        world = World()

        with world.modify_world():
            world.add_kinematic_structure_entity(bodies[0])
            for body in bodies:
                world.add_kinematic_structure_entity(body)
            joints = []
            for body_builder in factory.world_builder.body_builders:
                joints += self.parse_joints(body_builder=body_builder, world=world)
            for joint in joints:
                world.add_connection(joint)

        return world

    def parse_joints(self, body_builder: BodyBuilder, world: World) -> list[Connection]:
        """
        Parses joints from a BodyBuilder instance.
        :param body_builder: The BodyBuilder instance to parse.
        :param world: The World instance to add the connections to.
        :return: A list of Connection instances representing the parsed joints.
        """
        connections = []
        for joint_builder in body_builder.joint_builders:
            parent_body = world.get_kinematic_structure_entity_by_name(
                joint_builder.parent_prim.GetName()
            )
            child_body = world.get_kinematic_structure_entity_by_name(
                joint_builder.child_prim.GetName()
            )
            connection = self.parse_joint(joint_builder, parent_body, child_body, world)
            connections.append(connection)
        if (
            len(body_builder.joint_builders) == 0
            and not body_builder.xform.GetPrim().GetParent().IsPseudoRoot()
        ):
            parent_body = world.get_kinematic_structure_entity_by_name(
                body_builder.xform.GetPrim().GetParent().GetName()
            )
            child_body = world.get_kinematic_structure_entity_by_name(
                body_builder.xform.GetPrim().GetName()
            )
            transform = body_builder.xform.GetLocalTransformation()
            pos = transform.ExtractTranslation()
            quat = transform.ExtractRotationQuat()
            point_expr = cas.Point3(pos[0], pos[1], pos[2])
            quaternion_expr = cas.Quaternion(
                quat.GetImaginary()[0],
                quat.GetImaginary()[1],
                quat.GetImaginary()[2],
                quat.GetReal(),
            )
            origin = cas.TransformationMatrix.from_point_rotation_matrix(
                point=point_expr, rotation_matrix=quaternion_expr.to_rotation_matrix()
            )
            connection = FixedConnection(
                parent=parent_body,
                child=child_body,
                parent_T_connection_expression=origin,
            )
            connections.append(connection)

        return connections

    def parse_joint(
        self,
        joint_builder: JointBuilder,
        parent_body: Body,
        child_body: Body,
        world: World,
    ) -> Connection:
        joint_prim = joint_builder.joint.GetPrim()
        joint_name = joint_prim.GetName()
        joint_pos = joint_builder.pos
        joint_quat = joint_builder.quat
        point_expr = cas.Point3(joint_pos[0], joint_pos[1], joint_pos[2])
        quaternion_expr = cas.Quaternion(
            joint_quat.GetImaginary()[0],
            joint_quat.GetImaginary()[1],
            joint_quat.GetImaginary()[2],
            joint_quat.GetReal(),
        )
        origin = cas.TransformationMatrix.from_point_rotation_matrix(
            point=point_expr, rotation_matrix=quaternion_expr.to_rotation_matrix()
        )
        free_variable_name = joint_name
        offset = None
        multiplier = None
        if joint_prim.HasAPI(UsdUrdf.UrdfJointAPI):
            urdf_joint_api = UsdUrdf.UrdfJointAPI(joint_prim)
            if len(urdf_joint_api.GetJointRel().GetTargets()) > 0:
                free_variable_name = urdf_joint_api.GetJointRel().GetTargets()[0].name
                offset = urdf_joint_api.GetOffsetAttr().Get()
                multiplier = urdf_joint_api.GetMultiplierAttr().Get()
        if joint_builder.type == JointType.FREE:
            raise NotImplementedError("Free joints are not supported yet.")
        elif joint_builder.type == JointType.FIXED:
            return FixedConnection(
                parent=parent_body,
                child=child_body,
                parent_T_connection_expression=origin,
            )
        elif joint_builder.type in [
            JointType.REVOLUTE,
            JointType.CONTINUOUS,
            JointType.PRISMATIC,
        ]:
            axis = cas.Vector3(
                float(joint_builder.axis.to_array()[0]),
                float(joint_builder.axis.to_array()[1]),
                float(joint_builder.axis.to_array()[2]),
                reference_frame=parent_body,
            )
            try:
                dof = world.get_degree_of_freedom_by_name(free_variable_name)
            except KeyError:
                if joint_builder.type == JointType.CONTINUOUS:
                    lower_limits = DerivativeMap()
                    lower_limits.position = (
                        joint_builder.joint.GetLowerLimitAttr().Get()
                    )
                    upper_limits = DerivativeMap()
                    upper_limits.position = (
                        joint_builder.joint.GetUpperLimitAttr().Get()
                    )
                    dof = DegreeOfFreedom(
                        name=PrefixedName(joint_name),
                        lower_limits=lower_limits,
                        upper_limits=upper_limits,
                    )
                    world.add_degree_of_freedom(dof)
                else:
                    dof = DegreeOfFreedom(
                        name=PrefixedName(joint_name),
                    )
                    world.add_degree_of_freedom(dof)
            if joint_builder.type in [JointType.REVOLUTE, JointType.CONTINUOUS]:
                connection = RevoluteConnection(
                    parent=parent_body,
                    child=child_body,
                    parent_T_connection_expression=origin,
                    multiplier=multiplier,
                    offset=offset,
                    axis=axis,
                    dof_name=dof.name,
                )
            else:
                connection = PrismaticConnection(
                    parent=parent_body,
                    child=child_body,
                    parent_T_connection_expression=origin,
                    multiplier=multiplier,
                    offset=offset,
                    axis=axis,
                    dof_name=dof.name,
                )
            return connection
        else:
            raise NotImplementedError(
                f"Joint type {joint_builder.type} is not supported yet."
            )

    def parse_body(self, body_builder: BodyBuilder) -> Body:
        """
        Parses a body from a BodyBuilder instance.
        :param body_builder: The BodyBuilder instance to parse.
        :return: A Body instance representing the parsed body.
        """
        name = PrefixedName(
            prefix=self.prefix, name=body_builder.xform.GetPrim().GetName()
        )
        return Body(name=name)



import os
from collections import defaultdict

from typing_extensions import List

import numpy as np
import pytest
from rustworkx import NoPathFound

from semantic_digital_twin.reasoning.predicates import LeftOf
from semantic_digital_twin.robots.hsrb import HSRB
from semantic_digital_twin.spatial_types.spatial_types import TransformationMatrix
from semantic_digital_twin.world_description.connections import (
    OmniDrive,
    PrismaticConnection,
    RevoluteConnection,
)
from semantic_digital_twin.spatial_computations.ik_solver import (
    MaxIterationsException,
    UnreachableException,
)
from semantic_digital_twin.datastructures.prefixed_name import PrefixedName
from semantic_digital_twin.robots.abstract_robot import KinematicChain
from semantic_digital_twin.robots.tracy import Tracy
from semantic_digital_twin.robots.pr2 import PR2
from semantic_digital_twin.spatial_types.derivatives import Derivatives
from semantic_digital_twin.spatial_types.symbol_manager import symbol_manager
from semantic_digital_twin.world import World
from semantic_digital_twin.testing import pr2_world, tracy_world, hsrb_world


def test_compute_chain_of_bodies_pr2(pr2_world):
    root_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )
    tip_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    real = pr2_world.compute_chain_of_kinematic_structure_entities(
        root=root_link, tip=tip_link
    )
    real = [x.name for x in real]
    assert real == [
        PrefixedName(name="base_footprint", prefix="pr2"),
        PrefixedName(name="base_link", prefix="pr2"),
        PrefixedName(name="torso_lift_link", prefix="pr2"),
        PrefixedName(name="r_shoulder_pan_link", prefix="pr2"),
        PrefixedName(name="r_shoulder_lift_link", prefix="pr2"),
        PrefixedName(name="r_upper_arm_roll_link", prefix="pr2"),
        PrefixedName(name="r_upper_arm_link", prefix="pr2"),
        PrefixedName(name="r_elbow_flex_link", prefix="pr2"),
        PrefixedName(name="r_forearm_roll_link", prefix="pr2"),
        PrefixedName(name="r_forearm_link", prefix="pr2"),
        PrefixedName(name="r_wrist_flex_link", prefix="pr2"),
        PrefixedName(name="r_wrist_roll_link", prefix="pr2"),
        PrefixedName(name="r_gripper_palm_link", prefix="pr2"),
        PrefixedName(name="r_gripper_tool_frame", prefix="pr2"),
    ]


def test_compute_chain_of_connections_pr2(pr2_world):
    root_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )
    tip_link = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    real = pr2_world.compute_chain_of_connections(root=root_link, tip=tip_link)
    real = [x.name for x in real]
    assert real == [
        PrefixedName(name="base_footprint_joint", prefix="pr2"),
        PrefixedName(name="torso_lift_joint", prefix="pr2"),
        PrefixedName(name="r_shoulder_pan_joint", prefix="pr2"),
        PrefixedName(name="r_shoulder_lift_joint", prefix="pr2"),
        PrefixedName(name="r_upper_arm_roll_joint", prefix="pr2"),
        PrefixedName(name="r_upper_arm_joint", prefix="pr2"),
        PrefixedName(name="r_elbow_flex_joint", prefix="pr2"),
        PrefixedName(name="r_forearm_roll_joint", prefix="pr2"),
        PrefixedName(name="r_forearm_joint", prefix="pr2"),
        PrefixedName(name="r_wrist_flex_joint", prefix="pr2"),
        PrefixedName(name="r_wrist_roll_joint", prefix="pr2"),
        PrefixedName(name="r_gripper_palm_joint", prefix="pr2"),
        PrefixedName(name="r_gripper_tool_joint", prefix="pr2"),
    ]


def test_compute_chain_of_bodies_error_pr2(pr2_world):
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )
    with pytest.raises(NoPathFound):
        pr2_world.compute_chain_of_kinematic_structure_entities(root, tip)


def test_compute_chain_of_connections_error_pr2(pr2_world):
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )
    with pytest.raises(NoPathFound):
        pr2_world.compute_chain_of_connections(root, tip)


def test_compute_split_chain_of_bodies_pr2(pr2_world):
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_gripper_r_finger_tip_link")
    )
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_gripper_l_finger_tip_link")
    )
    chain1, connection, chain2 = (
        pr2_world.compute_split_chain_of_kinematic_structure_entities(root, tip)
    )
    chain1 = [n.name.name for n in chain1]
    connection = [n.name.name for n in connection]
    chain2 = [n.name.name for n in chain2]
    assert chain1 == [
        "l_gripper_r_finger_tip_link",
        "l_gripper_r_finger_link",
    ]
    assert connection == ["l_gripper_palm_link"]
    assert chain2 == ["l_gripper_l_finger_link", "l_gripper_l_finger_tip_link"]


def test_get_split_chain_pr2(pr2_world):
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_gripper_r_finger_tip_link")
    )
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_gripper_l_finger_tip_link")
    )
    chain1, chain2 = pr2_world.compute_split_chain_of_connections(root, tip)
    chain1 = [n.name.name for n in chain1]
    chain2 = [n.name.name for n in chain2]
    assert chain1 == ["l_gripper_r_finger_tip_joint", "l_gripper_r_finger_joint"]
    assert chain2 == ["l_gripper_l_finger_joint", "l_gripper_l_finger_tip_joint"]


def test_compute_fk_np_pr2(pr2_world):
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_gripper_tool_frame")
    )
    fk = pr2_world.compute_forward_kinematics_np(root, tip)
    np.testing.assert_array_almost_equal(
        fk,
        np.array(
            [
                [1.0, 0.0, 0.0, -0.0356],
                [0, 1.0, 0.0, -0.376],
                [0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )


def test_compute_fk_np_l_elbow_flex_joint_pr2(pr2_world):
    tip = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_elbow_flex_link")
    )
    root = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("l_upper_arm_link")
    )

    fk_expr = pr2_world.compose_forward_kinematics_expression(root, tip)
    fk_expr_compiled = fk_expr.compile()
    fk2 = fk_expr_compiled(
        symbol_manager.resolve_symbols(*fk_expr_compiled.symbol_parameters)
    )

    np.testing.assert_array_almost_equal(
        fk2,
        np.array(
            [
                [0.988771, 0.0, -0.149438, 0.4],
                [0.0, 1.0, 0.0, 0.0],
                [0.149438, 0.0, 0.988771, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ]
        ),
    )


def test_compute_ik(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[0, 3] -= 0.2
    joint_state = pr2_world.compute_inverse_kinematics(
        bf, eef, TransformationMatrix(fk, reference_frame=bf)
    )
    for joint, state in joint_state.items():
        pr2_world.state[joint.name].position = state
    pr2_world.notify_state_change()
    actual_fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    assert np.allclose(actual_fk, fk, atol=1e-3)


def test_compute_ik_max_iter(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("r_gripper_tool_frame")
    )
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[2, 3] = 10
    with pytest.raises(MaxIterationsException):
        pr2_world.compute_inverse_kinematics(
            bf, eef, TransformationMatrix(fk, reference_frame=bf)
        )


def test_compute_ik_unreachable(pr2_world):
    bf = pr2_world.root
    eef = pr2_world.get_kinematic_structure_entity_by_name(
        PrefixedName("base_footprint")
    )
    fk = pr2_world.compute_forward_kinematics_np(bf, eef)
    fk[2, 3] = -1
    with pytest.raises(UnreachableException):
        pr2_world.compute_inverse_kinematics(
            bf, eef, TransformationMatrix(fk, reference_frame=bf)
        )


def test_apply_control_commands_omni_drive_pr2(pr2_world):
    omni_drive: OmniDrive = pr2_world.get_connection_by_name(
        PrefixedName("odom_combined_T_base_footprint")
    )
    cmd = np.zeros((len(pr2_world.degrees_of_freedom)), dtype=float)
    cmd[pr2_world.state._index[omni_drive.x_velocity.name]] = 100
    cmd[pr2_world.state._index[omni_drive.y_velocity.name]] = 100
    cmd[pr2_world.state._index[omni_drive.yaw.name]] = 100
    dt = 0.1
    pr2_world.apply_control_commands(cmd, dt, Derivatives.jerk)
    assert pr2_world.state[omni_drive.yaw.name].jerk == 100.0
    assert pr2_world.state[omni_drive.yaw.name].acceleration == 100.0 * dt
    assert pr2_world.state[omni_drive.yaw.name].velocity == 100.0 * dt * dt
    assert pr2_world.state[omni_drive.yaw.name].position == 100.0 * dt * dt * dt

    assert pr2_world.state[omni_drive.x_velocity.name].jerk == 100.0
    assert pr2_world.state[omni_drive.x_velocity.name].acceleration == 100.0 * dt
    assert pr2_world.state[omni_drive.x_velocity.name].velocity == 100.0 * dt * dt
    assert pr2_world.state[omni_drive.x_velocity.name].position == 0

    assert pr2_world.state[omni_drive.y_velocity.name].jerk == 100.0
    assert pr2_world.state[omni_drive.y_velocity.name].acceleration == 100.0 * dt
    assert pr2_world.state[omni_drive.y_velocity.name].velocity == 100.0 * dt * dt
    assert pr2_world.state[omni_drive.y_velocity.name].position == 0

    assert pr2_world.state[omni_drive.x.name].jerk == 0.0
    assert pr2_world.state[omni_drive.x.name].acceleration == 0.0
    assert pr2_world.state[omni_drive.x.name].velocity == 0.0
    assert pr2_world.state[omni_drive.x.name].position == 0.08951707486311977

    assert pr2_world.state[omni_drive.y.name].jerk == 0.0
    assert pr2_world.state[omni_drive.y.name].acceleration == 0.0
    assert pr2_world.state[omni_drive.y.name].velocity == 0.0
    assert pr2_world.state[omni_drive.y.name].position == 0.1094837581924854


def test_search_for_connections_of_type(pr2_world: World):
    connections = pr2_world.get_connections_by_type(OmniDrive)
    assert len(connections) == 1
    assert connections[0].name == PrefixedName(
        name="odom_combined_T_base_footprint", prefix="pr2"
    )
    assert connections[0].parent == pr2_world.root
    assert connections[0].child == pr2_world.get_kinematic_structure_entity_by_name(
        "base_footprint"
    )

    connections = pr2_world.get_connections_by_type(PrismaticConnection)
    assert len(connections) == 3
    assert connections[0].name == PrefixedName(name="torso_lift_joint", prefix="pr2")
    assert connections[0].parent == pr2_world.get_kinematic_structure_entity_by_name(
        "base_link"
    )
    assert connections[0].child == pr2_world.get_kinematic_structure_entity_by_name(
        "torso_lift_link"
    )
    assert connections[1].name == PrefixedName(
        name="r_gripper_motor_slider_joint", prefix="pr2"
    )
    assert connections[1].parent == pr2_world.get_kinematic_structure_entity_by_name(
        "r_gripper_palm_link"
    )
    assert connections[1].child == pr2_world.get_kinematic_structure_entity_by_name(
        "r_gripper_motor_slider_link"
    )
    assert connections[2].name == PrefixedName(
        name="l_gripper_motor_slider_joint", prefix="pr2"
    )
    assert connections[2].parent == pr2_world.get_kinematic_structure_entity_by_name(
        "l_gripper_palm_link"
    )
    assert connections[2].child == pr2_world.get_kinematic_structure_entity_by_name(
        "l_gripper_motor_slider_link"
    )

    connections = pr2_world.get_connections_by_type(RevoluteConnection)
    assert len(connections) == 40


def test_pr2_semantic_annotation(pr2_world):
    pr2 = PR2.from_world(pr2_world)

    # Ensure there are no loose bodies
    pr2_world._notify_model_change()

    assert len(pr2.manipulators) == 2
    assert len(pr2.manipulator_chains) == 2
    assert len(pr2.sensors) == 1
    assert len(pr2.sensor_chains) == 1
    assert pr2.neck == list(pr2.sensor_chains)[0]
    assert pr2.torso.name.name == "torso"
    assert len(pr2.torso.sensors) == 0
    assert list(pr2.sensor_chains)[0].sensors == pr2.sensors
    assert pr2.left_arm and pr2.right_arm
    assert pr2.left_arm != pr2.right_arm


def test_specifies_left_right_arm_mixin(pr2_world):
    pr2 = PR2.from_world(pr2_world)
    left_arm_chain = list(pr2.left_arm.bodies)
    right_arm_chain = list(pr2.right_arm.bodies)
    assert LeftOf(
        left_arm_chain[1],
        right_arm_chain[1],
        pr2.root.global_pose,
    )()


def test_kinematic_chains(pr2_world):
    pr2 = PR2.from_world(pr2_world)
    semantic_kinematic_chain_annotation: List[KinematicChain] = (
        pr2_world.get_semantic_annotations_by_type(KinematicChain)
    )
    for chain in semantic_kinematic_chain_annotation:
        assert chain.root
        assert chain.tip


def test_load_collision_config_srdf(pr2_world):
    path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "..",
        "resources",
        "collision_configs",
        "pr2.srdf",
    )
    pr2_world.load_collision_srdf(path)
    assert len([b for b in pr2_world.bodies if b.get_collision_config().disabled]) == 20
    assert len(pr2_world.disabled_collision_pairs) == 1128


def test_tracy_semantic_annotation(tracy_world):
    tracy = Tracy.from_world(tracy_world)

    tracy_world._notify_model_change()

    assert len(tracy.manipulators) == 2
    assert len(tracy.manipulator_chains) == 2
    assert len(tracy.sensors) == 1
    assert len(tracy.sensor_chains) == 1
    assert tracy.torso is None
    assert list(tracy.sensor_chains)[0].sensors == tracy.sensors


def test_hsrb_semantic_annotation(hsrb_world):
    hsrb = HSRB.from_world(hsrb_world)

    hsrb_world._notify_model_change()

    assert len(hsrb.manipulators) == 1
    assert len(hsrb.manipulator_chains) == 1
    assert hsrb.neck is not None
    assert len(hsrb.arms) == 1

    assert len(hsrb.sensors) == 5
    assert len(hsrb.sensor_chains) == 2
    assert hsrb.torso is not None


def test_pr2_tighten_dof_velocity_limits_of_1dof_connections(pr2_world):
    pr2 = PR2.from_world(pr2_world)

    # set all joints to vel limit 1
    pr2.tighten_dof_velocity_limits_of_1dof_connections(defaultdict(lambda: 1))

    # try spacial case for specific joint
    new_limits = defaultdict(
        lambda: 0.5, {pr2._world.get_connection_by_name("head_pan_joint"): 23}
    )
    pr2.tighten_dof_velocity_limits_of_1dof_connections(new_limits)
    # if spacial case triggers, but the new limit is above the old one, nothing happens
    assert (
        pr2._world.get_connection_by_name("head_pan_joint").dof.upper_limits.velocity
        == 1
    )
    # new limit is applied to joint without spacial case
    assert (
        pr2._world.get_connection_by_name("head_tilt_joint").dof.upper_limits.velocity
        == 0.5
    )
    # non-spacial case where the old limit is below 1
    assert (
        pr2._world.get_connection_by_name("torso_lift_joint").dof.upper_limits.velocity
        == 0.013
    )
